#ifndef __FIND_CONTOUR_H
#define __FIND_CONTOUR_H
#include "meshelt.h"
#include "fixme.h"
#include <list>

template<class Mesh_T, class Interpolant_T>
class Contour : public Mesh_T::TraitsType
  {
public:
    typedef Mesh_T MeshType;
    typedef Interpolant_T InterpolantType;
    typedef InterpolantType::ContourType ContourType;
    typedef std::list<ContourType> ContourList;

    InterpolantType mInterpolant;
    MeshType mMesh;
    MeshElement<MeshType> mElement;
    ScalarType mEpsilon;
    ScalarType mEpsilonRatio;
    ScalarType mEpsilonInterpolation;
    // mEpsilon is divided up into:
    // * approximation error (mEpsilonRatio * mEpsilon)
    // * interpolation error ((1-mEpsilonRatio) * mEpsilon)

    Contour() : mEpsilonRatio(0.9)
      { }

    void Init()
      { // Assumes working set initialized to domain
        // Clear result set
        while (mMesh.PopResult()  != mMesh.InvalidElement())
          ;
        // initialize mesh member of element
        mElement.mMesh=&mMesh;
        mInterpolant.mEpsilon=mEpsilon*(1-mEpsilonRatio);
        mInterpolant.mEpsilonR=mInterpolant.mEpsilon*
                            Matrix::Min(.25, .5*mInterpolant.mEpsilon);
        mEpsilonInterpolation=mEpsilon*mEpsilonRatio;
      }
    bool Step();
    void Results(ContourList &results);
  };

template<class Mesh_T, class Function_T>
bool Contour<Mesh_T, Function_T>::Step()
  {
    // while working set not empty
    //   current element = pop element from working set
    MeshType::ElementType element=mMesh.PopWorking();
    if (element==mMesh.InvalidElement())
      // if working set empty, we are done
      return false;

    // evaluate each corner that is not already evaluated
    int c;
    for (c=0; c<mMesh.NumCorners(element); ++c)
      mInterpolant.EvaluateCorner(mMesh.Corner(element, c));

    ScalarType elementradius=mMesh.Radius(element);

    // if element small, stop refining, and put in result set
    if (elementradius*2 <= mEpsilonInterpolation)
      {
        mMesh.PushResult(element);
        return true;
      }

    mMesh.AllocateCenter(element);
    MeshType::VertexType *center=& mMesh.Center(element);
    mInterpolant.EvaluateCenter(*center);

    mElement.mElement = element;
    ElementAction action=mInterpolant.
                  AnalyzeElement(&mElement, elementradius, 
                                  center, mEpsilonInterpolation);
    switch (action)
      {
        case RefineElement:
          mMesh.Refine(element);
          break;
        case DiscardElement:
          mMesh.DiscardElement(element);
          break;
        case ContourElement:
          mMesh.PushResult(element);
          break;
        default:
          assert(0 && "should not get here");
      }
    return true;
  }

template<class Mesh_T, class Function_T>
void Contour<Mesh_T, Function_T>::Results(ContourList &results)
  {
    MeshElementList<Mesh_T> mesh;
    mesh.Init(&mMesh);
    mInterpolant.InterpolateContour(results, &mesh);
  }

#endif