/* to become contourmd.h */
#ifndef __FIND_CONTOUR_MAXD_H
#define __FIND_CONTOUR_MAXD_H
#include "meshelt.h"
#include <list>

template<class Mesh_T, class Interpolant_T>
class ContourMaxD : public Mesh_T::TraitsType
  {
public:
    typedef Mesh_T MeshType;
    typedef Interpolant_T InterpolantType;
    typedef InterpolantType::ContourType ContourType;
    typedef std::list<ContourType> ContourList;

    InterpolantType mInterpolant;
    MeshType mMesh;
    MeshElement<MeshType> mElement;
    ScalarType mEpsilon;
    // mEpsilon is divided up into:
    // * approximation error (cEpsilonRatio * mEpsilon)
    // * interpolation error ((1-cEpsilonRatio) * mEpsilon)

    static const ScalarType cEpsilonRatio;

    void Init()
      { // Assumes working set initialized to domain
        // Clear result set
        while (mMesh.PopResult()  != mMesh.InvalidElement())
          ;
        // initialize mesh member of element
        mElement.mMesh=&mMesh;
        mInterpolant.mEpsilon=mEpsilon*(1-cEpsilonRatio);
        mInterpolant.mEpsilonR=mInterpolant.mEpsilon*
                            Matrix::Min(.25, .5*mInterpolant.mEpsilon);
      }
    bool Step();
    void Results(ContourList &results);
  };

template<class Mesh_T, class Interpolant_T>
  const ScalarType 
  ContourMaxD<Mesh_T, Interpolant_T>::cEpsilonRatio = 0.9;

template<class Mesh_T, class Interpolant_T>
bool ContourMaxD<Mesh_T, Interpolant_T>::Step()
  {
    // while working set not empty
    //   current element = pop element from working set
    MeshType::ElementType element=mMesh.PopWorking();
    if (element==mMesh.InvalidElement())
      // if working set empty, we are done
      return false;

    ScalarType elementradius=mMesh.Radius(element);

    // if whole element covered by != 0 region of any evaluated corner:
    //   discard this element
    if (mMesh.IsCoveredByCorners(element, elementradius))
      {
        mMesh.DiscardElement(element);
        return true;
      }

    // evaluate each corner that is not already evaluated
    int c;
    for (c=0; c<mMesh.NumCorners(element); ++c)
      if (! (mMesh.Corner(element, c).mEvaluated & MeshType::VertexType::ValueEvaluated))
        mInterpolant.EvaluateValue(mMesh.Corner(element, c));

    // if element small, stop refining, and put in result set
    if (elementradius*2 <= mEpsilon*cEpsilonRatio)
      {
        mMesh.PushResult(element);
        return true;
      }

    // if refined enough
    //    add to result set
    mElement.mElement = element;
    if (mInterpolant.StoppingCriteria(&mElement, elementradius, mEpsilon*cEpsilonRatio))
      {
        mMesh.PushResult(element);
        return true;
      }

    // if signs differ:
    //   refine
    bool pos=mMesh.Corner(element, 0).mFunctionValue>0;
    for (c=1; c<mMesh.NumCorners(element); ++c)
      if ((mMesh.Corner(element, c).mFunctionValue>0)!=pos)
        {
          mMesh.Refine(element);
          return true;
        }

    //--- check center ---
    //allocate center vertex
    //get radius needed to cover element
    //ask function if that radius is safe
    //if so, discard
    //else refine
    mMesh.AllocateCenter(element);
    MeshType::VertexType *center=& mMesh.Center(element);
    ScalarType targetradius = mMesh.RadiusForCenterToCover(element,elementradius);
    assert(center->mSafeRadius==0);
    center->mSafeRadius = mInterpolant.SafeRadius((*center), targetradius);
    if (center->mSafeRadius >= targetradius)
      {
        mMesh.DiscardElement(element);
      }
    else
      {
        mMesh.Refine(element);
      }
    return true;
  }

template<class Mesh_T, class Interpolant_T>
void ContourMaxD<Mesh_T, Interpolant_T>::Results(ContourList &results)
  {
    MeshElementList<Mesh_T> mesh;
    mesh.Init(&mMesh);
    mInterpolant.InterpolateContour(results, &mesh);
  }

#endif