#ifndef __DENSE_MATRIX_H
#define __DENSE_MATRIX_H

namespace Matrix {

// Class DenseRow<N> ------------------------------------
// N is either float, double, Complex, etc. (Number)
// Must be Initialized with a size before being used.

template <class N>
class DenseColumn;

template <class N>
class DenseRow
  {
protected:    
    N *mElts;
    int mW;

    void DoCopy(const DenseRow<N> &m); // assumes reset has been called

    // Do we support any operations for compatibility with DenseMatrix below?
    // Probably not, all our functions use hefty doses of template magic
public:
    typedef N Number;

    DenseRow() : mElts(0), mW(0) { }
    DenseRow(const DenseRow<N> & cpy) { DoCopy(cpy); }

    void Initialize(int w)
      {
        assert(mElts==0);
        mW=w;
        mElts=new N[mW];
      }

    typedef DenseColumn<N> ColumnType;  
    void InitializeColumn(ColumnType &c) const;

    typedef DenseRow<N> RowType;
    void InitializeRow(RowType &r) const
      { r.Initialize(mW); }

    typedef DenseColumn<N> TransposeType;
    void InitializeTranspose(TransposeType &t) const;

    void Reset()
      {
        if (mElts)
          {
            delete[] mElts;
            mElts=0;
            mW=0;
          }
      }
    ~DenseRow() { Reset(); }

    DenseRow<N> & operator=(const DenseRow<N> & cpy) 
      { Reset(); DoCopy(cpy); return *this; }

    N operator()(int r, int c) const
      { assert(mElts); 
        assert(r==0);
        assert(c>=0 && c<mW);
        return mElts[c];
      }
    N & operator()(int r, int c)
      { assert(mElts); 
        assert(r==0);
        assert(c>=0 && c<mW);
        return mElts[c];
      }

    int Width() const { return mW; }
    int Height() const { return 1; }
  };

template<class N>
inline void DenseRow<N>::DoCopy(const DenseRow<N> &m)
  { 
    mW = m.Width();
    mElts = new N[mW];
    for (int c=0; c<mW; ++c)
      (*this)(0,c) = m(0,c);
  }


// Class DenseColumn<N> ---------------------------------
// N is either float, double, Complex, etc. (Number)
// Must be Initialized with a size before being used.
template <class N>
class DenseColumn
  {
protected:    
    N *mElts;
    int mH;

    void DoCopy(const DenseColumn<N> &m);

    // Do we support any operations for compatibility with DenseMatrix below?
    // Probably not, all our functions use hefty doses of template magic
public:
    typedef N Number;

    DenseColumn() : mElts(0), mH(0) { }
    DenseColumn(const DenseColumn<N> & cpy) { DoCopy(cpy); }

    void Initialize(int h)
      {
        assert(mElts==0);
        mH=h;
        mElts=new N[mH];
      }

    typedef DenseColumn<N> ColumnType;  
    void InitializeColumn(ColumnType &c) const
      {
        c.Initialize(mH);
      }


    typedef DenseRow<N> RowType;
    void InitializeRow(RowType &r) const
      {
        r.Initialize(1);
      }

    typedef DenseRow<N> TransposeType;
    void InitializeTranspose(TransposeType &t) const
      {
        t.Initialize(mH);
      }

    void Reset()
      {
        if (mElts)
          {
            delete[] mElts;
            mElts=0;
            mH=0;
          }
      }
    ~DenseColumn() { Reset(); }

    DenseColumn<N> & operator=(const DenseColumn<N> & cpy) 
      { Reset(); DoCopy(cpy); return *this; }

    N operator()(int r, int c) const
      { assert(mElts); 
        assert(c==0);
        assert(r>=0 && r<mH);
        return mElts[r];
      }
    N & operator()(int r, int c)
      { assert(mElts); 
        assert(c==0);
        assert(r>=0 && r<mH);
        return mElts[r];
      }

    int Width() const { return 1; }
    int Height() const { return mH; }
  };

template<class N>
inline void DenseColumn<N>::DoCopy(const DenseColumn<N> &m)
  { // assumes reset has been called
    mH = m.Height();
    mElts = new N[mH];
    for (int r=0; r<mH; ++r)
      (*this)(r,0) = m(r,0);
  }

// DenseRow methods which refer to DenseColumn members

template<class N>
inline void DenseRow<N>::InitializeColumn(ColumnType &c) const
  { c.Initialize(1); }

template<class N>
inline void DenseRow<N>::InitializeTranspose(TransposeType &t) const
  { t.Initialize(mW); }

// Class DenseMatrix<N> ---------------------------------
// N is either float, double, Complex, etc. (Number)
// Must be Initialized with a size before being used.
template <class N>
class DenseMatrix
  {
protected:    
    N *mElts;
    int mH, mW;

    void DoCopy(const DenseMatrix<N> &m)
      { // assumes reset has been called
        mH = m.Height();
        mW = m.Width();
        mElts = new N[mH*mW];
        for (int r=0; r<mH; ++r)
          for (int c=0; c<mW; ++c)
            {
              (*this)(r,c) = m(r,c);
            }
      }
    /* Do member templates work on anything?
    template<class M>
    void DoCopy(const M &m)
      { // assumes reset has been called
        mH = m.Height();
        mW = m.Width();
        mElts = new N[mH*mW];
        for (int r=0; r<mH; ++r)
          for (int c=0; c<mW; ++c)
            {
              (*this)(r,c) = m(r,c);
            }
      }
    */
public:
    typedef N Number;

    DenseMatrix() : mElts(0), mW(0), mH(0) { }
    DenseMatrix(const DenseMatrix<N> & cpy) { DoCopy(cpy); }

    void Initialize(int h, int w)
      {
        assert(mElts==0);
        mH=h;
        mW=w;
        mElts=new N[mW*mH];
      }

    typedef DenseColumn<N> ColumnType;  
    void InitializeColumn(ColumnType &c) const
      {
        c.Initialize(mH);
      }

    typedef DenseRow<N> RowType;
    void InitializeRow(RowType &r) const
      {
        r.Initialize(mW);
      }

    typedef DenseMatrix<N> TransposeType;
    void InitializeTranspose(TransposeType &t) const
      {
        t.Initialize(mW, mH);
      }

    void Reset()
      {
        if (mElts)
          {
            delete[] mElts;
            mElts=0;
            mH=0;
            mW=0;
          }
      }
    ~DenseMatrix() { Reset(); }

    DenseMatrix<N> & operator=(const DenseMatrix<N> & cpy) 
      { Reset(); DoCopy(cpy); return *this; }

    N operator()(int r, int c) const
      { assert(mElts); 
        assert(r>=0 && r<mH);
        assert(c>=0 && c<mW);
        return mElts[r*mW+c];
      }
    N & operator()(int r, int c)
      { assert(mElts); 
        assert(r>=0 && r<mH);
        assert(c>=0 && c<mW);
        return mElts[r*mW+c];
      }

    int Width() const { return mW; }
    int Height() const { return mH; }
  };

}

#endif