#ifndef __ROOT_OF_SPLINE_2D_H
#error findcntr.hpp should only be included from root2d.hpp
#endif

#include "fixme.h"
#include <map>
#include "smallmat.h"
#include "matrix.h"

// compute curvatures to use de Boor et al's method of G1 Hermite interpolation
#define CURVATURE
#define N_ROOTFINDER

namespace Root2D
  {
    // MISC HELPER FUNCTIONS for FindContour(...) ---------------------

    class ConstTCubicFinder
      {
    protected:
        Root1D::BezierUnit  mLeft;
        Root1D::BezierUnit  mRight;
        ScalarType mSPartial[5];
    public:
        void Initialize(const TriangularCubicUnit  &cubic);
        void CubicFromT(ScalarType t, Root1D::BezierInterval 
                         & output) const;
      };

    
    class Cubic2dContourHelper
      {
    protected:
        // Types
        typedef TriangularQuadraticUnit  QuadraticType;
        typedef ConstTCubicFinder  CubicFinderType;
        typedef Root1D::BezierUnit  Cubic1dUnit;
        typedef Root1D::BezierInterval  Cubic1dInterval;
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> STLocation;

        typedef enum { NoTransition, StartAtEdge, EndAtEdge, Max, Min, 
          Isolated, Horizontal, HorizontalAtEdge, Unknown, UnknownAtEdge, 
          DeletePoint } TransitionType;

    public:
        struct TransitionPoint
          {
            ScalarType mS, mT;
            TransitionType mTransition;
            int mBeforeAfter[2];
            ScalarType mNormal[2];
#ifdef CURVATURE
            ScalarType mCurvature;
#endif
            int mSide;
          };

        /*            Before      After
        NoTransition    1           1
        StartAtEdge     0           1
        EndAtEdge       1           0
        Max             2           0
        Min             0           2
        Isolated        0           0
        Horizontal, HorizontalAtEdge, Unknown, UnknownAtEdge
                        ?           ?
        */


        typedef TransitionPoint  TransitionPointType;
        typedef Contour ContourType;

    protected:
        typedef std::vector<TransitionPointType>   TransitionPointList;
        typedef TransitionPointList::iterator      TransitionPointIter;
        struct TransitionRow
          {
            ScalarType mLastT;
            TransitionPointList mTrans;
          };
        typedef std::map<ScalarType, TransitionRow> TDataList;
        typedef TDataList::iterator                 TDataIter;

        struct Panel
          {
            TDataIter beginend[2];
            TDataIter guidelines[2];
            ScalarType sintersections[2][3];
            int numlines;
          };

        typedef std::list<Panel> PanelList;
        typedef PanelList::iterator PanelIter;

        struct Strand
          {

            PanelIter mPanelSpan[2];
            std::vector<int> mStrandNum;
            std::vector<int> mTurningNum;
            TransitionPointType *mEndPts[2];
            int mConnectedTo[2];
            std::vector<STLocation> mControlPolygon;
          };

        typedef Strand StrandType;

        struct StrandCollection
          {
            typedef std::vector<StrandType> StrandStorage;
            typedef StrandStorage::iterator StrandIter;
            StrandStorage mStrands;
            std::vector<int> mFrontier;
            void CreateOneStrand(TransitionPointType *tp, 
                                  PanelIter panel, int num, int turningpoint);
            void CreateTwoStrands(TransitionPointType *tp, 
                                  PanelIter panel, int num, int turningpoint);
            void StarSize(int first, int num, int which);
            void EndSegment();
            void EndOneStrand(TransitionPointType *tp, PanelIter panel, 
                              int which);
            void JoinTwoStrands(TransitionPointType *tp, 
                                PanelIter panel, int first);
          };

    public:
        struct StarData
          {
            ScalarType mS;
            int mNum;
          };

    protected:
        struct SplinePoint
          {
            STLocation mST;
            STLocation mTangent;
#ifdef CURVATURE
            ScalarType mCurvature;
#endif
          };

        struct Pass1Data
          {
            SplinePoint mPoint;
            int mTotalRoots;
            int mRootNum;
            TransitionType mTransition; 
            // only care about Unknown, UnknownAtEdge, and Horizontal
          };

    protected:
        // Data
        enum { WorstSmallEpsilon=196 };
        ScalarType mEpsilon, mEpsilonR;
        CubicFinderType mCubicFinder;
        TDataList mTData;
        QuadraticType mGradient[2];

    protected:
        // Helper functions
        //typedef enum { Zero, Negative, Positive } SignResult;
        typedef Root1D::SignResult SignResult;
        enum { Zero=Root1D::Zero, 
               Negative=Root1D::Negative, 
               Positive=Root1D::Positive
             };

        static SignResult SignOfProduct(ScalarType a, ScalarType b)
          { return Root1D::SignOfProduct(a,b); }

        // uses mEpsilon, mEpsilonR
        TransitionType ClassifyEdgePoint(TransitionPointType *point, 
              ScalarType secondpartials[2][2], bool zero[2], int side, 
              bool iscorner);

        static inline bool TransitionPair(TransitionType a, TransitionType b, 
                                          TransitionType c, TransitionType d);

        // uses mTData
        TransitionPointList * FindLine(ScalarType t, ScalarType epsilon);

        // uses mTData
        TransitionPointType * FindDuplicate(ScalarType s, ScalarType t, 
                                            ScalarType epsilon);



        // uses mGradient, mEpsilon
        void AnalyzeInteriorPoint(const ScalarType st[2], 
                                  TransitionPointType *point);

        // uses mGradient, mEpsilonR, mCubicFinder, mGradient
        void RegularLine(ScalarType t, ScalarType epsilon, 
                          TransitionRow &output);

        static bool MergePoints(TransitionPoint &, TransitionPoint &);

#ifdef CURVATURE
        static void FindCircleCenter(STLocation &center, ScalarType radius, 
                       const ScalarType normal[2]);
        static ScalarType ComputeCurvature(ScalarType gradient[2], 
                                    ScalarType secondpartials[2][2]);

        static inline ScalarType Cross2d(const STLocation & v, 
                                         const STLocation & w);

        static void RescaleQuadratic(QuadraticType &quad);

        // uses mEpsilon, mEpsilonR
        bool FindControlPoints(const SplinePoint *inpts[2],
                               STLocation controlpts[2]);

        // uses mEpsilon, mEpsilonR
        ScalarType NthRoot(const Cubic1dInterval &cubic, int rootnum, 
                           int totalnumroots);

        // uses mCubicFinder, mGradient, mEpsilon, mEpsilonR
        void OptimizeSplinePath(int rootnum, int totalnumroots,
                                const SplinePoint *pts[2],
                                std::vector<STLocation> &results);

        // uses mEpsilon, mEpsilonR
        void SimpleJoin(const SplinePoint *ptrs[2],
                        STLocation controlpts[2]);

        // uses mEpsilon, mEpsilonR
        void MinMaxJoin(const TransitionPointType *inpts[2], ScalarType t,
                        bool increasingt[2], STLocation controlpts[2]);
#else
        // uses mEpsilon, mEpsilonR
        void FindControlPoints(const SplinePoint *inpts[3],
                                STLocation controlpts[2]);

        // uses mCubicFinder, mGradient, mEpsilon, mEpsilonR
        void OptimizeSplinePath(int rootnum, int totalnumroots, 
                                const SplinePoint pts[3], 
                                std::vector<STLocation> &results);
#endif

        static int RotateSide(int rotatecounterclockwise, int torotate);

        void TransitionToSplinePoint(const TransitionPoint &tp, 
                                    SplinePoint &sp);

        void UnknownToKnown(const SplinePoint &u, const SplinePoint &k, 
                     STLocation &out1, STLocation &out2);

      };

    
    inline bool operator<(const Cubic2dContourHelper ::
                            TransitionPointType &left, 
                          const Cubic2dContourHelper ::
                            TransitionPointType &right)
      { return left.mS < right.mS; }

    
    inline bool operator<(const Cubic2dContourHelper ::StarData &left,
                   const Cubic2dContourHelper ::StarData &right)
      {
        return left.mS < right.mS;
      }

    
    inline bool Cubic2dContourHelper 
      ::TransitionPair(TransitionType a, TransitionType b, 
                               TransitionType c, TransitionType d)
      {
        return (a==c && b==d) || (a==d && b==c);
      }

    template <class _T1, class _T2>
    inline std::pair<const _T1, _T2> make_map_pair(const _T1& __x, 
                                                   const _T2& __y)
      {
        return std::pair<const _T1, _T2>(__x, __y);
      }

    class Cubic2dContourFinder : public Cubic2dContourHelper 
      {
    protected:
        // Types
        typedef TriangularCubic  CubicType;
        typedef std::list<ScalarType> RootList;
        typedef RootList::iterator RootIter;

    protected:
        // Data
        CubicType mCubic;
        RootList mRootsOnSides[3];
        bool mZeroSides[3];
        bool mBadSides[3];
        ScalarType mBadT[3];
        int mGoodSide;
        PanelList mPanels;
        StrandCollection mStrands;
        std::list<ContourType> *mResults;

    protected:
        // Helper functions, use mCubic
        void OutputVertex(std::list<ContourType> & outto, 
                          const STLocation &st);
        STLocation PushLineSegmentResult(std::list<ContourType> & outto, 
                            const STLocation &st, ScalarType s, ScalarType t);
        void OutputStraightLineSegment(std::list<ContourType> & outto,
                                const STLocation &st1, const STLocation &st2, 
                                int side1, int side2);

        void ConnectEdgePoints(const TransitionPoint &a, 
                               const TransitionPoint &b);

        // Steps of the algorithm
        void FactorOutZeroSides();
        void FindRootsOnEachSide();
        void IncrementGoodSide();
        int  DetermineGoodSide();
        void ThreeStraightLinesCase();
        void ComputeGradients();
        void InsertEdgePoints();
        void InsertInteriorPoints();
        void RemoveDeletedPoints();
        void FindRootsAtSameT();
        void RemoveRedundantInterior();
        void InitializePanels();
        bool NoPanelsCase();
        void ComputeGuideLines();
        bool SetBeforeAfter();
        bool BuildStrands();
        void SplineForStrands();
        void StitchConnectedStrands();
        void OutputZeroSides();
        bool TryAnOrientation();

    public:
        Cubic2dContourFinder(const CubicType &cubic, ScalarType epsilon, 
                      ScalarType epsilonR, std::list<ContourType> & results);
      };
  }

