#include "root2d.h"
#include "findcntr.h"

namespace Root2D
  {
    void ConstTCubicFinder::Initialize(const TriangularCubicUnit  &cubic)
      {
        cubic.CubicFromSide(1, mRight);
        cubic.CubicFromSide(2, mLeft);
        std::swap(mLeft.mY[0], mLeft.mY[3]);
        std::swap(mLeft.mY[1], mLeft.mY[2]);

        TriangularQuadraticUnit  spartial;
        cubic.PartialDerivativeS(spartial);
        mSPartial[0]=spartial.mY[0];
        mSPartial[1]=spartial.mY[3];
        mSPartial[2]=spartial.mY[5];
        mSPartial[3]=spartial.mY[4];
        mSPartial[4]=spartial.mY[2];
      }

    
    void ConstTCubicFinder::CubicFromT(ScalarType t, Root1D::BezierInterval 
                     & output) const
      {
        output.mX0=0;
        output.mXd=1-t;
        ScalarType scale=(1-t)/3;
        output.mY[0]=mLeft.EvaluateAt(t);
        output.mY[3]=mRight.EvaluateAt(t);
        ScalarType tmp0, tmp1;
        tmp0 = mSPartial[0]+(mSPartial[1]-mSPartial[0])*t;
        tmp1 = mSPartial[1]+(mSPartial[2]-mSPartial[1])*t;
        tmp0 = tmp0+(tmp1-tmp0)*t;
        output.mY[1]=output.mY[0]+tmp0*scale;
        tmp0 = mSPartial[4]+(mSPartial[3]-mSPartial[4])*t;
        tmp1 = mSPartial[3]+(mSPartial[2]-mSPartial[3])*t;
        tmp0 = tmp0+(tmp1-tmp0)*t;
        output.mY[2]=output.mY[3]-tmp0*scale;
      }

    /* side
            2 |\ 1
               0
     */
    
    Cubic2dContourHelper ::TransitionType 
    Cubic2dContourHelper 
    ::ClassifyEdgePoint(TransitionPoint  *point,
              ScalarType secondpartials[2][2], bool zero[2], int side, 
              bool iscorner)
      {
        // determine type of transition based on if components of 
        // normal are nonzero and "side" the side the root is on (as
        // opposed to "side", the side of the triangle we are using
        // as a base)
        if (! zero[0])
          {
            if (side)
              // may be EndAtEdge or NoTransition
              if (side==1)
                {
                  ScalarType diff = point->mNormal[1]-
                                       point->mNormal[0];
                  bool transverse = Matrix::Abs(diff)>mEpsilonR;
                  if (! transverse)
                    return NoTransition;
                  else if (SignOfProduct(diff, point->mNormal[0])
                                  == Negative)
                    return EndAtEdge;
                }
              else // side==2
                {
                  if (zero[1])  // if zero, not transverse
                    return NoTransition;
                  else if (SignOfProduct(point->mNormal[0], 
                                     point->mNormal[1]) == Positive)
                    return EndAtEdge;
                }
            return StartAtEdge;
          }
        else if (! zero[1])
          { // tangent to boundary case, find curvature
            SignResult sign = SignOfProduct(point->mNormal[1], 
                                            secondpartials[0][0]);
            if (sign == Positive) // local max
              {
                if (side)
                  return EndAtEdge;
                else // side==0
                  // we could mark this point as isolated, 
                  // but we will discard it since it will
                  // be included in the adjacent triangle
                  return Isolated;
              }
            else if (sign == Negative) // local min
              return side ? StartAtEdge : Min;
              // if Min, should we duplicate this point?
            else if (side || iscorner)
              return HorizontalAtEdge;
            else
              return Horizontal;
          }
        else
          { // gradient=0 case
            ScalarType dethessian = 
                secondpartials[0][0]*secondpartials[1][1] -
                secondpartials[0][1]*secondpartials[1][0];
            if (dethessian>0)
              { // isolated point, local minimum of function
                return Isolated;
              }
            else 
              { // crossing! what to do!?
                return UnknownAtEdge;
              }
          }
        assert(0);
      }

    
    Cubic2dContourHelper ::TransitionPointList *
    Cubic2dContourHelper 
    ::FindLine(ScalarType t, ScalarType epsilon)
      {
        TDataIter cur = mTData.lower_bound(t);
        TDataIter end = mTData.upper_bound(t+epsilon);
        if (cur==end)
          {
            cur=mTData.insert(cur, make_map_pair(t, TransitionRow()));
            (*cur).second.mLastT = t;
            return & (*cur).second.mTrans;
          }
        (*cur).second.mLastT=Matrix::Max(t, (*cur).second.mLastT);
        return & (*cur).second.mTrans;
      }

    
    Cubic2dContourHelper ::TransitionPointType *
    Cubic2dContourHelper 
    ::FindDuplicate(ScalarType s, ScalarType t, ScalarType epsilon)
      {
        TDataIter cur = mTData.lower_bound(t-epsilon);
        TDataIter end = mTData.upper_bound(t+epsilon);
        while (cur!=end)
          {
            ScalarType tdiff=Matrix::Abs((*cur).first-t);
            for (TransitionPointIter tpi=(*cur).second.mTrans.begin();
                 tpi!=(*cur).second.mTrans.end(); ++tpi)
              {
                if (Matrix::Abs((*tpi).mS-s)+tdiff < epsilon)
                  return &(*tpi);
              }
            ++cur;
          }
        return 0;
      }

#ifdef CURVATURE

    
    void Cubic2dContourHelper 
    ::FindCircleCenter(STLocation &center, ScalarType radius, 
                       const ScalarType normal[2])
      {
        ScalarType len=sqrt(normal[0]*normal[0]+normal[1]*normal[1]);
        radius/=len;
        center(0,0) += radius*normal[0];
        center(0,1) += radius*normal[1];
      }

    
    ScalarType Cubic2dContourHelper 
    ::ComputeCurvature(ScalarType gradient[2], ScalarType secondpartials[2][2])
      {
        // Input: gradients and second partials of function f at point a
        // f(a) is assumed to b 0
        assert(Matrix::Abs(gradient[0])>0.0000001 || 
               Matrix::Abs(gradient[1])>0.0000001);
        ScalarType normsq = gradient[0]*gradient[0]+gradient[1]*gradient[1];
        // Output: curvature of zero set of f at point a

        // Fact: curvature of the parabola y=kx^2 at the vertex (0,0) is 2k
        // Locally the contour is a (rotated) parabola,
        // we want to figure out "k"
        // f(x+a) is approximately f(a)+Dot(gradient,x)+
        //                              .5*transpose(x)*curvature*x
        // so f(x+a)=0 when Dot(gradient,x)+.5*transpose(x)*curvature*x=0
        // the vertex of parabola is at a (x=0)
        // the gradient is perpendicular to the parabola at the vertex
        // for curvature actually only care about second partial in 
        // the direction perpendicular to the gradient (along the parabola's
        // tangent direction)
        // Let tangent = perpendicular(gradient)
        // (note |tangent| == |gradient|)
        // Let unittangent = tangent / |tangent|
        // Let secondorderterm = .5*transpose(unittangent) * secondpartials * 
        //                          unittangent
        // ( = transpose(tangent) * secondpartials * tangent / |tangent|^2 )
        // Let s = Dot(x,tangent)/|tangent|
        // Let t = Dot(x,gradient)/|gradient|
        // So s,t is a local ("gauge") coordinate system for f aligned 
        // with the parabola, then:
        // f(x+a) is approximately secondorderterm*s^2 + |gradient|*t
        //  + k1 * s * t + k2 * t^2
        // setting this to 0 and solving for t:
        // secondorderterm*s^2 + (|gradient|+k1*s)*t + k2 * t^2 = 0
        // a s^2 + b*t + k2 t^2 = 0

        // t = [-b +/- sqrt(b^2 - 4*a*s^2*k2)]/(2*k2)
        // know t=0 when s=0 (and b=|gradient|), so want + branch
        // t = [-b + sqrt(b^2 - 4*a*s^2*k2)]/(2*k2)
        // multiplying and dividing by the conjugate:
        // t = [b^2 - b^2 + 4*a*s^2*k2]/(2*k2*[-b-sqrt(b^2 - 4*a*s^2*k2)]
        // t = 4*a*s^2*k2/(2*k2*[-b-sqrt(b^2 - 4*a*s^2*k2)]
        // t = 2*a*s^2/[-b-sqrt(b^2 - 4*a*s^2*k2)]
        // near s=0 this is approximately
        // t = 2*a*s^2/[-2|gradient|]
        // t = - secondorderterm * s^2 / |gradient|
        // So k = - secondorderterm / |gradient|
        // Or k = -.5*transpose(tangent)*secondpartials*tangent / |gradient|^3
        // And curvature is twice that, so drop the .5
        ScalarType numer=(-gradient[1]*secondpartials[0][0]+
                          gradient[0]*secondpartials[0][1])*gradient[1]-
                         (-gradient[1]*secondpartials[1][0]+
                          gradient[0]*secondpartials[1][1])*gradient[0];
        return numer/sqrt(normsq*normsq*normsq);
      }
#endif

    
    void Cubic2dContourHelper 
    ::AnalyzeInteriorPoint(const ScalarType st[2],
                  TransitionPoint  *point)
      {
        ScalarType secondpartials[2][2];
        mGradient[0].EvaluateAt(st, point->mNormal[0],
                                secondpartials[0]);
        mGradient[1].EvaluateAt(st, point->mNormal[1],
                                secondpartials[1]);
        // bool dtiszero = Matrix::Abs(point->mNormal[1])<mEpsilon;
        bool dtiszero = NearAZero(mGradient[1], st, mEpsilon, mEpsilonR);

        // the point will be within epsilon (in the domain)
        // of where the x gradient is zero, but the 
        // x gradient may be bigger than epsilon
        // if slope of gradient is locally greater than 1
        //assert(Matrix::Abs(point->mNormal[0])<mEpsilon);
        point->mSide = -1;
        if (dtiszero)
          { // gradient 0 case
            point->mTransition=Unknown;
          }
        else
          {
            SignResult sign = SignOfProduct(point->mNormal[1], 
                                            secondpartials[0][0]);
#ifdef CURVATURE
            if (sign == Zero)
              {
                point->mTransition=Horizontal;
                point->mCurvature=0;
              }
            else
              {
                if (sign == Positive)
                  point->mTransition=Max;
                else 
                  {
                    assert(sign == Negative);
                    point->mTransition=Min;
                  }
                point->mCurvature=ComputeCurvature(point->mNormal, 
                                                  secondpartials);
              }
#else
            if (sign == Positive)
              point->mTransition=Max;
            else if (sign == Negative)
              point->mTransition=Min;
            else
              point->mTransition=Horizontal;
#endif
          }
      }

    
    void Cubic2dContourHelper 
    ::RegularLine(ScalarType t, ScalarType epsilon, 
                  TransitionRow &output)
      {
        output.mLastT=t;

        Root1D::BezierInterval  b;
        mCubicFinder.CubicFromT(t, b);
        typedef std::list<ScalarType> RootList;
        typedef RootList::iterator RootIter;
        RootList roots;
#ifndef N_ROOTFINDER
        Root1D::Solve_Quadratic(roots, b, epsilon, epsilonR);
#else
        Root1D::Solve_N_P(roots, b, epsilon, mEpsilonR);
#endif
        assert(roots.size()<=3);
        roots.sort();
        for (RootIter riter = roots.begin();
             riter != roots.end(); ++riter)
          {
            // (compute the normal vector, as well)
            TransitionPoint  toadd;
            toadd.mS=(*riter);
            toadd.mT=t;
            toadd.mTransition=NoTransition;
            STPoint  st;
            st.st[0]=(*riter);
            st.st[1]=t;
#ifdef CURVATURE
            ScalarType secondpartials[2][2];
            mGradient[0].EvaluateAt(st.st, toadd.mNormal[0], secondpartials[0]);
            mGradient[1].EvaluateAt(st.st, toadd.mNormal[1], secondpartials[1]);
            assert(Matrix::Abs(toadd.mNormal[0])>mEpsilonR);
            toadd.mCurvature=ComputeCurvature(toadd.mNormal, secondpartials);
#else
            toadd.mNormal[0]=gradient[0].EvaluateAt(st.st);
            toadd.mNormal[1]=gradient[1].EvaluateAt(st.st);
            assert(Matrix::Abs(toadd.mNormal[0])>mEpsilonR);
#endif
            toadd.mSide = -1;
            output.mTrans.push_back(toadd);
          }
      }

    
    bool Cubic2dContourHelper 
    ::MergePoints(TransitionPoint &a, TransitionPoint &b)
      {
        TransitionType at=a.mTransition,
                       bt=b.mTransition;
        if (at==NoTransition)
          {
            a=b;
          }
        else if (bt==NoTransition)
          {
            b=a;
          }
        else if (a.mSide!=-1 && b.mSide!=-1 && a.mSide!=b.mSide)
          return false;
        else if (at==StartAtEdge || at==EndAtEdge || at==UnknownAtEdge)
          {
            a.mTransition=UnknownAtEdge;
            b=a;
          }
        else if (bt==StartAtEdge || bt==EndAtEdge || bt==UnknownAtEdge)
          {
            b.mTransition=UnknownAtEdge;
            a=b;
          }
        else
          {
            a.mTransition=Unknown;
            a.mS = (a.mS+b.mS)*.5;
            a.mT = (a.mT+b.mT)*.5;
            b=a;
          }
        return true;
      }


     
    void Cubic2dContourHelper ::StrandCollection
    ::CreateOneStrand(TransitionPointType *tp, PanelIter panel, int num, 
                      int turningpoint)
      {
        int strandindex = mStrands.size();
        mStrands.push_back();
        mStrands.back().mPanelSpan[0]=panel;
        mStrands.back().mEndPts[0]=tp;
        mStrands.back().mConnectedTo[0]= -1;
        mStrands.back().mTurningNum.push_back(turningpoint);
        assert(num>=0);
        assert(num<=mFrontier.size());
        mFrontier.insert(mFrontier.begin()+num, strandindex);
      }

    
    void Cubic2dContourHelper ::StrandCollection
    ::CreateTwoStrands(TransitionPointType *tp, PanelIter panel, int num, 
                      int turningpoint)
      {
        int firststrandindex = mStrands.size();
        mStrands.push_back();
        mStrands.back().mPanelSpan[0]=panel;
        mStrands.back().mEndPts[0]=tp;
        mStrands.back().mConnectedTo[0]=firststrandindex+1;
        mStrands.back().mTurningNum.push_back(turningpoint);
        mStrands.push_back();
        mStrands.back().mPanelSpan[0]=panel;
        mStrands.back().mEndPts[0]=tp;
        mStrands.back().mConnectedTo[0]=firststrandindex;
        mStrands.back().mTurningNum.push_back(turningpoint);
        assert(num>=0);
        assert(num<=mFrontier.size());
        mFrontier.insert(mFrontier.begin()+num, firststrandindex);
        mFrontier.insert(mFrontier.begin()+num+1, firststrandindex+1);
      }

    
    void Cubic2dContourHelper ::StrandCollection
    ::StarSize(int first, int num, int which)
      {
        for (int i=first; i<first+num; ++i)
          {
            assert(i<mFrontier.size());
            assert(mFrontier[i]<mStrands.size());
            mStrands[mFrontier[i]].mTurningNum.push_back(which);
          }
      }

    
    void Cubic2dContourHelper ::StrandCollection
    ::EndSegment()
      {
        assert(mFrontier.size()<=3);
        for (int i=0; i<mFrontier.size(); ++i)
          {
            assert(mFrontier[i]<mStrands.size());
            mStrands[mFrontier[i]].mStrandNum.push_back(i);
          }
      }

    
    void Cubic2dContourHelper ::StrandCollection
    ::EndOneStrand(TransitionPointType *tp, PanelIter panel, int which)
      {
        assert(which>=0);
        assert(which<mFrontier.size());
        StrandType *strand = & mStrands[mFrontier[which]];
        strand->mPanelSpan[1]=panel;
        strand->mEndPts[1]=tp;
        strand->mConnectedTo[1]= -1;
        mFrontier.erase(mFrontier.begin()+which);
      }

    
    void Cubic2dContourHelper ::StrandCollection
    ::JoinTwoStrands(TransitionPointType *tp, PanelIter panel, int first)
      {
        assert(first>=0);
        assert(first+1<mFrontier.size());
        StrandType *strand[2] = { & mStrands[mFrontier[first]],
                                  & mStrands[mFrontier[first+1]] };
        strand[0]->mPanelSpan[1]=panel;
        strand[0]->mEndPts[1]=tp;
        strand[0]->mConnectedTo[1]= mFrontier[first+1];
        strand[1]->mPanelSpan[1]=panel;
        strand[1]->mEndPts[1]=tp;
        strand[1]->mConnectedTo[1]= mFrontier[first];
        mFrontier.erase(mFrontier.begin()+first, 
                        mFrontier.begin()+first+2);
      }

#ifdef CURVATURE
    
    inline ScalarType Cubic2dContourHelper 
    ::Cross2d(const STLocation & v, const STLocation & w)
      {
        return v(0,0)*w(0,1)-v(0,1)*w(0,0);
      }

    
    void Cubic2dContourHelper 
    ::RescaleQuadratic(QuadraticType &quad)
      {
        ScalarType maxval=0;
        int i;
        for (i=0; i<6; ++i)
          {
            ScalarType absval=Matrix::Abs(quad.mY[i]);
            if (absval>maxval)
              maxval=absval;
          }
        if (maxval<1 && maxval>0)
          {
            for (i=0; i<6; ++i)
              quad.mY[i]/=maxval;
          }
      }

    
    bool Cubic2dContourHelper 
    ::FindControlPoints(const SplinePoint *inpts[2], STLocation controlpts[2])
      {
        // TANGENTS HAVE TO BE NORMALIZED
        assert(Matrix::Abs(Matrix::TwoNorm(inpts[0]->mTangent)-1)<0.01);
        assert(Matrix::Abs(Matrix::TwoNorm(inpts[1]->mTangent)-1)<0.01);

        // aliases tangents[0,1] == d0,1
        // curvatures[0,1] == k0,1
        // verts[0,1] == f0,1 == b0,3
        // controlpts[0,1] = b1,2

        // v x w := v[0]*w[1]-v[1]*w[0]  ; see Cross2d above

        // want b1 and b2 where
        // b1 = inpts[0]->mST + delta0*inpts[0]->mTangent
        // b2 = inpts[1]->mST - delta1*inpts[1]->mTangent
        // delta0,1 > 0
        // ideally inpts[0]->mCurvature = 2 inpts[0]->mTangent x (b2-b1)/(3 delta0^2)
        // and     inpts[1]->mCurvature = 2 inpts[1]->mTangent x (b1-b2)/(3 delta1^2)
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> STLocation;
        STLocation a;
        // a = inpts[1]->mST-inpts[0]->mST
        Matrix::Subtract(a, inpts[1]->mST, inpts[0]->mST);
        // want (inpts[0]->mTangent x inpts[1]->mTangent) delta0 = (a x inpts[1]->mTangent) -
        //        (3/2)*inpts[1]->mCurvature*delta1^2
        // and  (inpts[0]->mTangent x inpts[1]->mTangent) delta1 = (inpts[0]->mTangent x a) -
        //        (3/2)*inpts[0]->mCurvature*delta0^2
        ScalarType d0xd1 = Cross2d(inpts[0]->mTangent, inpts[1]->mTangent);
        ScalarType dxa[2];
        dxa[0]=Cross2d(inpts[0]->mTangent, a); // d0xa
        dxa[1]=Cross2d(a, inpts[1]->mTangent); // axd1

        // d0xd1 * delta0 = dxa[1] - (3/2)*inpts[1]->mCurvature*delta1^2
        // d0xd1 * delta1 = dxa[0] - (3/2)*inpts[0]->mCurvature*delta0^2

        // deal with special case: curvature==0
        {
          ScalarType delta[2];
          bool foundsolution=false;
          if (Matrix::Abs(d0xd1)>mEpsilonR)
            {
              if (Matrix::Abs(inpts[0]->mCurvature)<mEpsilonR)
                {
                  // d0xd1*delta1 = dxa[0]
                  // d0xd1*delta0 = dxa[1] - (3/2)*inpts[1]->mCurvature*delta1^2
                  delta[1] = dxa[0]/d0xd1;
                  delta[0] = (dxa[1]-1.5*inpts[1]->mCurvature*delta[1]*delta[1])/d0xd1;
                  foundsolution=true;
                }
              else if (Matrix::Abs(inpts[1]->mCurvature)<mEpsilonR)
                {
                  // d0xd1 * delta0 = dxa[1] 
                  // d0xd1 * delta1 = dxa[0] - (3/2)*inpts[0]->mCurvature*delta0^2
                  delta[0] = dxa[1]/d0xd1;
                  delta[1] = (dxa[0]-1.5*inpts[0]->mCurvature*delta[0]*delta[0])/d0xd1;
                  foundsolution=true;
                }
            }
          else
            // if (Matrix::Abs(inpts[0]->mCurvature)<mEpsilonR || 
            //       Matrix::Abs(inpts[1]->mCurvature)<mEpsilonR)
            { // could be linear
              // dxa[i] = (3/2)*inpts[i]->mCurvature*delta[i]^2
              for (int i=0; i<2; ++i)
                if (Matrix::Abs(inpts[i]->mCurvature)<mEpsilonR)
                  if (Matrix::Abs(dxa[i])<mEpsilonR)
                    // linear case, any delta works (condition is 0=0)
                    delta[i] = Matrix::TwoNorm(a)/3;
                  else
                    // inconsistent
                    return false;
                else 
                  { 
                    ScalarType quotient=dxa[i]/inpts[i]->mCurvature;
                    if (quotient<0)
                      return false;
                    delta[i] = sqrt((2./3)*quotient);
                  }
              foundsolution=true;
            }
          if (foundsolution)
            {
              if (delta[0]<0 || delta[1]<0)
                return false;
              controlpts[0]=inpts[0]->mST;
              Matrix::AddScaled(controlpts[0], delta[0], inpts[0]->mTangent);
              controlpts[1]=inpts[1]->mST;
              Matrix::AddScaled(controlpts[1], -delta[1], inpts[1]->mTangent);
              return true;
            }
        }

        ScalarType extents[2];
        bool extents_exists[2]={false,false};
        int i;
        for (i=0; i<2; ++i)
          if (SignOfProduct(d0xd1, inpts[i]->mCurvature)!=Negative)
            {
              if (SignOfProduct(inpts[i]->mCurvature, dxa[i])==Negative)
                return false;
              extents[i]=2*sqrt((2./3)*dxa[i]/inpts[i]->mCurvature);
              extents_exists[i] = true;
            }

        if (!extents_exists[0] && !extents_exists[1])
          {
            // each of the equations:
            // d0xd1 * delta0 = dxa[1] - (3/2)*inpts[1]->mCurvature*delta1^2
            // d0xd1 * delta1 = dxa[0] - (3/2)*inpts[0]->mCurvature*delta0^2
            // defines a parabola in the delta0, delta1 plane
            // (and we are looking for intersections in the first quadrant)
            // extents_exists[i] if the parabola is bounded in the
            // first quadrant, so in this case both are unbounded
            // Consider the line delta0=delta1:
            // If neither equation intersects that line, there is no 
            // intersection, since they must be on opposite sides of the line.
            // Otherwise, there can be no intersection after the last
            // intersection with the line, so we have a bound on the solutions
            // a=(3/2)*inpts[i]->mCurvature
            // b=d0xd1
            // c=-dxa[i]
            bool foundroot=false;
            for (i=0; i<2; ++i)
              {
                ScalarType descriminant=d0xd1*d0xd1+
                                        6*inpts[i]->mCurvature*dxa[i];
                if (descriminant>=0)
                  {
                    ScalarType root = (-d0xd1+sqrt(descriminant))/
                                      (3*inpts[i]->mCurvature);
                    if (! foundroot || root>extents[0])
                      extents[0]=root;
                    foundroot=true;
                  }
              }
            if (! foundroot || extents[0]<0)
              return false;
            extents[1]=extents[0];
          }
        else
          { // at least one extent exists
            // use it to improve the other extent
            if (Matrix::Abs(d0xd1)>mEpsilonR)
              {
                for (i=0; i<2; ++i)
                  if (extents_exists[i])
                    {
                      ScalarType oppextent=2*dxa[i]/d0xd1;
                      // improve existing or initialize non-existant extent
                      if (extents_exists[1-i])
                        extents[1-i] = Matrix::Min(extents[1-i],oppextent);
                      else
                        extents[1-i] = oppextent;
                    }
              }
            else 
              {
                assert(extents_exists[i] && extents_exists[i]);
              }
          }

        // rewrite as f=0, g=0
        // d0xd1 * delta0 - dxa[1] + (3/2)*inpts[1]->mCurvature*delta1^2 = 0
        // d0xd1 * delta1 - dxa[0] + (3/2)*inpts[0]->mCurvature*delta0^2 = 0
        // d0xd1 * delta0 - dxa[1] + (3/2)*inpts[1]->mCurvature*delta1^2 = 0
        // d0xd1 * delta1 = dxa[0] - (3/2)*inpts[0]->mCurvature*delta0^2
        // solving for delta0:
        // d0xd1 * delta0 - dxa[1] + (3/2)*inpts[1]->mCurvature*
        // (dxa[0] - (3/2)*inpts[0]->mCurvature*delta0^2)^2 / d0xd1^2 = 0

        typedef STPoint  PointType;
        typedef std::list<PointType> ResultsType;
        typedef ResultsType::const_iterator ResultsCIter;
        ResultsType results;

          {
            typedef Root1D::BezierNthDegreeUnit  
                    GeneralBezier;
            // quartic = dxa[0]/d0xd1 - ((3/2)*inpts[0]->mCurvature/d0xd1)*t^2
            // for t=0 .. extents[0]

            // A = dxa[0]/d0xd1
            // B = (3/2)*inpts[0]->mCurvature*d0xd1
            // C = (3/2)*inpts[1]->mCurvature
            // D = d0xd1
            // E = dxa[1]
            // quartic = dxa[0]/d0xd1 = A
            GeneralBezier quartic(2, dxa[0]/d0xd1);
            // quadratic = t*B
            // (t=0..extents[0])
            GeneralBezier quadratic(1, 0, 1.5*inpts[0]->mCurvature*extents[0]/
                                          d0xd1);
            // quadratic *= t  (t=0..extents[0])
            // so quadratic = t*t*B
            quadratic *= GeneralBezier(1, 0, extents[0]);
            // quartic = A - t*t*B
            quartic -= quadratic;
            // quartic = (A - t*t*B)^2
            quartic *= quartic;
            // quartic = C*(A - t*t*B)^2
            quartic *= 1.5*inpts[1]->mCurvature;
            // eqn = d0xd1*t-dxa[1] == D*t-E
            GeneralBezier eqn(4, -dxa[1], d0xd1*extents[0]-dxa[1]);
            // eqn = D*t- E + C*(A - t*t*B)^2
            eqn += quartic;
            // now eqn = f(extents[0]*t) for t = 0 ... 1
            // where f(x) = d0xd1*x-dxa[1] + (3/2)*inpts[1]->mCurvature*
            //    (dxa[0]/d0xd1 - x*x*(3/2)*inpts[0]->mCurvature*d0xd1)^2
            eqn.Rescale();
            ScalarType x[2];
            x[0]=0;
            x[1]=1;
            typedef std::list<ScalarType> Results0Type;
            typedef Results0Type::const_iterator ResultsC0Iter;
            Results0Type results0;
            Root1D::Solve_N(results0, 4, eqn.GetOrdinates(), x, 
                            mEpsilon/extents[0], mEpsilonR);

            ResultsC0Iter iter;
            // are there any non-negative solutions?
            for (iter=results0.begin(); iter!=results0.end(); ++iter)
              {
                PointType st;
                st.st[0]=*iter;
                if (st.st[0]>=0)
                  {
                    ScalarType temp=st.st[0]*extents[0];
                    st.st[1] = dxa[0] - 1.5*inpts[0]->mCurvature*temp*temp;
                    st.st[1] /= d0xd1;

                    //assert(st.st[1]<=extents[1]*(1+mEpsilon));

                    if (st.st[1]>=0)
                      {
                        st.st[0]*=extents[0];
                        results.push_back(st);
                      }
                  }
              }
          }

        if (results.empty())
          // no non-negative solutions
          return false;

        ResultsCIter smallest=results.begin();

        // search for *smallest* non-negative solution
        ScalarType smallestsize=(*smallest).st[0] + (*smallest).st[1];
        ResultsCIter compareto=smallest;
        for (++compareto; compareto!=results.end(); ++compareto)
          if ((*compareto).st[0] >= 0 && (*compareto).st[1] >= 0)
            {
              ScalarType comparesize = (*compareto).st[0] + (*compareto).st[1];
              if (comparesize<smallestsize)
                {
                  smallest=compareto;
                  smallestsize=comparesize;
                }
            }

        // found smallest solution, use it to set controlpts
        controlpts[0]=inpts[0]->mST;
        Matrix::AddScaled(controlpts[0], (*smallest).st[0], inpts[0]->mTangent);
        controlpts[1]=inpts[1]->mST;
        Matrix::AddScaled(controlpts[1], -(*smallest).st[1], inpts[1]->mTangent);

        return true;
      }

    
    ScalarType Cubic2dContourHelper 
    ::NthRoot(const Cubic1dInterval &cubic, int rootnum, int totalnumroots)
      {
        assert(rootnum<totalnumroots);
        assert(rootnum>=0);
        ScalarType small_epsilon=mEpsilon;
        typedef std::list<ScalarType> RootsType;
        typedef RootsType::iterator RootsIter;
        RootsType roots;
        do
          {
            roots.clear();
#ifndef N_ROOTFINDER
            Root1D::Solve_Quadratic(roots, cubic, small_epsilon, mEpsilonR);
#else
            Root1D::Solve_N_P(roots, cubic, small_epsilon, mEpsilonR);
#endif
            assert(roots.size()<=3);
            if (roots.size()==totalnumroots)
              {
                roots.sort();
                RootsIter rootiter=roots.begin();
                std::advance(rootiter, rootnum);
                return *rootiter;
              }
            small_epsilon *= .5;
            // should epsilonR *= .25 here?
          } while (small_epsilon > mEpsilon/WorstSmallEpsilon);

        roots.sort();
        if (roots.size() > totalnumroots)
          { // should merge closest until down to correct number
            do
              { // toatlnumroots>=1 so roots.size()>=2
                assert(roots.size()>=2);
                RootsIter first=roots.begin();
                assert(first!=roots.end());
                RootsIter second=first;
                ++second;
                assert(second!=roots.end());
                RootsIter closest=first;
                ScalarType dist=(*second)-(*first);
                for (first=second, ++second; second!=roots.end(); 
                     first=second, ++second)
                  {
                    ScalarType newdist = (*second)-(*first);
                    if (newdist<dist)
                      {
                        closest=first;
                        dist=newdist;
                      }
                  }
                // merge closest and closest+1
                *closest += dist*.5;
                ++closest;
                roots.erase(closest);
              } while (roots.size() > totalnumroots);
            // fall through to return root
          }
        else // too few roots, do the best we can
          {
            if (roots.size() <= rootnum)
              if (roots.empty())
                return cubic.mXd*.5+cubic.mX0;
              else
                return roots.back();
            // else fall through to return root
          }
        assert(roots.size() > rootnum);
        RootsIter rootiter=roots.begin();
        std::advance(rootiter, rootnum);
        return *rootiter;
      }

    
    void Cubic2dContourHelper 
    ::OptimizeSplinePath(int rootnum, int totalnumroots, 
                         const SplinePoint *pts[2],
                         std::vector<STLocation> &results)
      {
        // --- Algorithm ---
        // keep a list of points with: s, t, unit tangent, curvature
        // while list has at least two elements:
        //   form cubic spline connecting first two
        //   (if fails, have to subdivide at average of t's, see below)
        //   (also have to subdivide if d/dt <=0 ever)
        //   find s,t of "midpoint" (parametrically) of cubic spline
        //   use cubicfinder to find roots at that t
        //   (should check to make sure have right number of roots)
        //   compare midpoint s to s of rootnum-th root
        //   if difference bigger than epsilon,
        //     "subdivide"
        //     compute gradient, unit tangent, and curvature for root
        //     insert it into the list between first two points
        //   otherwise,
        //     output control points to results
        //     remove first point from list
        //   (note that it won't subdivide if the spline points ever
        //    get within epsilon distance)
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> STLocation;
        typedef SplinePoint  Point;
        typedef std::list<Point> PointList;
        typedef PointList::iterator PointIter;
        PointList points;
        points.push_back(*pts[0]);
        points.push_back(*pts[1]);
        while (points.size()>1)
          {
            const Point *firsttwo[2];
            firsttwo[0] = & points.front();
            firsttwo[1] = & *(++points.begin());
            STLocation ctrlpts[2];
            bool subdivide;
            STLocation midpt;
            if (! FindControlPoints(firsttwo, ctrlpts))
              { // use a simple interpolation scheme that does not
                // depend on curvatures
                ScalarType length = Matrix::Length(firsttwo[0]->mST, firsttwo[1]->mST);
                // ctrlpts[0] = firsttwo[0]->mST + length*firsttwo->mTangent/3;
                ctrlpts[0] = firsttwo[0]->mST;
                Matrix::AddScaled(ctrlpts[0], length/3, firsttwo[0]->mTangent);
                ctrlpts[1] = firsttwo[1]->mST;
                Matrix::AddScaled(ctrlpts[1], -length/3, firsttwo[1]->mTangent);
              }
            // check resulting spline increases monotonically
            ScalarType deltat[3];
            deltat[0]=ctrlpts[0](0,1)-firsttwo[0]->mST(0,1);
            deltat[1]=ctrlpts[1](0,1)-ctrlpts[0](0,1);
            deltat[2]=firsttwo[1]->mST(0,1)-ctrlpts[1](0,1);
            if (deltat[0]>=0 && deltat[2]>=0 && deltat[1]<0)
              {
                ScalarType seconddiff[2];
                seconddiff[0]=deltat[1]-deltat[0]; // <0
                seconddiff[1]=deltat[2]-deltat[1]; // >0
                ScalarType minat = -seconddiff[0]/
                                      (seconddiff[1]-seconddiff[0]);
                seconddiff[0]=deltat[0]+(deltat[1]-deltat[0])*minat;
                seconddiff[1]=deltat[1]+(deltat[2]-deltat[1])*minat;
                ScalarType minval = seconddiff[0]+
                    (seconddiff[1]-seconddiff[0])*minat;

                if (minval<=0)
                  { // does not increase monotonically, need to shorten 
                    // derivatives
                    // this scalefact should be conservative
                    ScalarType totalt=firsttwo[1]->mST(0,1)-
                            firsttwo[0]->mST(0,1);
                    ScalarType scalefact= deltat[1]/(deltat[1]-totalt);
                    Matrix::AddScaledDiff(ctrlpts[0], scalefact,
                              firsttwo[0]->mST, ctrlpts[0]);
                    Matrix::AddScaledDiff(ctrlpts[1], scalefact,
                              firsttwo[1]->mST, ctrlpts[1]);
                  }
              }
            // compute (parametric-)mid-point of spline
            // == first/8 + second*3/8 + third*3/8 + fourth/8
            Matrix::Add(midpt, firsttwo[0]->mST, firsttwo[1]->mST);
            Matrix::AddScaled(midpt, 3, ctrlpts[0]);
            Matrix::AddScaled(midpt, 3, ctrlpts[1]);
            Matrix::Scale(midpt, .125);
            subdivide=false;

            Cubic1dInterval b;
            mCubicFinder.CubicFromT(midpt(0,1), b);
            ScalarType actuals=NthRoot(b, rootnum, totalnumroots);
            if (Matrix::Abs(actuals-midpt(0,0))>mEpsilon)
              {
                // maybe we can choose a better point on
                // the bezier curve and get closer to the actual point
                STLocation derivative;
                Matrix::Subtract(derivative, firsttwo[1]->mST, 
                                             firsttwo[0]->mST);
                Matrix::Subtract(derivative, derivative, ctrlpts[0]);
                Matrix::Add(derivative, derivative, ctrlpts[1]);
                Matrix::Scale(derivative, .75);
                ScalarType normsq = derivative(0,0)*derivative(0,0)+
                                    derivative(0,1)*derivative(0,1);
                ScalarType newt = .5 + (actuals-midpt(0,0))*
                                       derivative(0,0)/normsq;
                
                // have a new t value, plug into our current bezier
                // will get a new point to compare to point on contour
                midpt(0,0)=actuals; // midpt is now on contour
                STLocation temp[6];
                temp[0]=firsttwo[0]->mST;
                Matrix::AddScaledDiff(temp[0], newt, 
                      ctrlpts[0], firsttwo[0]->mST);
                temp[1]=ctrlpts[0];
                Matrix::AddScaledDiff(temp[1], newt, 
                      ctrlpts[1], ctrlpts[0]);
                temp[2]=ctrlpts[1];
                Matrix::AddScaledDiff(temp[2], newt, 
                      firsttwo[1]->mST, ctrlpts[1]);
                temp[3]=temp[0];
                Matrix::AddScaledDiff(temp[3], newt, temp[1], temp[0]);
                temp[4]=temp[1];
                Matrix::AddScaledDiff(temp[4], newt, temp[2], temp[1]);
                temp[5]=temp[3];
                Matrix::AddScaledDiff(temp[5], newt, temp[4], temp[3]);

                ScalarType length=Matrix::Length(midpt, temp[5]);
                if (length>mEpsilon)
                  subdivide=true;
              }
            if (subdivide)
              { // if two points are within epsilon, don't bother
                // subdividing
                ScalarType length=Matrix::Length(firsttwo[0]->mST, 
                                                 firsttwo[1]->mST);
                if (length<mEpsilon)
                  {
                    subdivide=false;
                    ScalarType maxtdelta=firsttwo[1]->mST(0,1)-
                                          firsttwo[0]->mST(0,1);
                    for (int i=0; i<2; ++i)
                      { // set ctrlpts[i]
                        ScalarType tangentlen=length/3;
                        ScalarType tdelta=tangentlen*
                                            firsttwo[0]->mTangent(0,1);
                        if (tdelta>maxtdelta)
                          {
                            tangentlen *= maxtdelta/tdelta;
                            tdelta=maxtdelta;
                          }
                        ScalarType sdelta=tangentlen*
                                            firsttwo[0]->mTangent(0,0);
                        ctrlpts[i]=firsttwo[i]->mST;
                        if (i)
                          {
                            ctrlpts[i](0,0)-=sdelta;
                            ctrlpts[i](0,1)-=tdelta;
                          }
                        else
                          {
                            ctrlpts[i](0,0)+=sdelta;
                            ctrlpts[i](0,1)+=tdelta;
                          }
                      }
                  }
              }
            if (subdivide)
              {
                // add new point to be processed
                // figure out gradient and curvature at midpt
                Point newpt;
                newpt.mST=midpt;
                ScalarType position[2];
                position[0]=midpt(0,0);
                position[1]=midpt(0,1);
                ScalarType normal[2];
                ScalarType secondpartials[2][2];
                mGradient[0].EvaluateAt(position, normal[0], secondpartials[0]);
                mGradient[1].EvaluateAt(position, normal[1], secondpartials[1]);

                newpt.mCurvature = ComputeCurvature(normal, secondpartials);
                // ensure t component of tangent is positive
                if (normal[0]>0)
                  {
                    newpt.mTangent(0,0) = -normal[1];
                    newpt.mTangent(0,1) = normal[0];
                    // normal to right of tangent, so flip sign of curvature
                    newpt.mCurvature = -newpt.mCurvature;
                  }
                else
                  {
                    newpt.mTangent(0,0) = normal[1];
                    newpt.mTangent(0,1) = -normal[0];
                  }
                // normalization below is somewhat redundant with
                // computations in ComputeCurvature above
                ScalarType tangentlen=Matrix::TwoNorm(newpt.mTangent);
                Matrix::Scale(newpt.mTangent, 1/tangentlen);
                points.insert(++points.begin(), newpt);
              }
            else
              { // output three control points
                results.push_back(firsttwo[0]->mST);
                results.push_back(ctrlpts[0]);
                results.push_back(ctrlpts[1]);
                // remove point from list of points to process
                points.pop_front();
              }
          }
      }

    
    void Cubic2dContourHelper 
    ::SimpleJoin(const SplinePoint *ptrs[2], STLocation controlpts[2])
      {
        if (!FindControlPoints(ptrs, controlpts))
          { // a simpler join, no chance of failure, does not use curvature
            STLocation direct;
            Matrix::Subtract(direct, ptrs[1]->mST, ptrs[0]->mST);
            ScalarType l1=Matrix::Dot(ptrs[0]->mTangent, direct)/3;
            ScalarType l2=Matrix::Dot(ptrs[1]->mTangent, direct)/3;
            assert(l1>=0);
            assert(l2>=0);
            controlpts[0]=ptrs[0]->mST;
            controlpts[1]=ptrs[1]->mST;
            Matrix::AddScaled(controlpts[0], l1, ptrs[0]->mTangent);
            Matrix::AddScaled(controlpts[1], -l2, ptrs[1]->mTangent);
          }
      }
  
    
    void Cubic2dContourHelper 
    ::MinMaxJoin(const TransitionPoint *inpts[2], ScalarType t, 
                 bool increasingt[2], STLocation controlpts[2])
      {
        assert(inpts[0]->mTransition==NoTransition);
        assert(inpts[1]->mTransition==NoTransition);
        assert(Matrix::Abs(inpts[0]->mNormal[0])>mEpsilonR);
        assert(Matrix::Abs(inpts[1]->mNormal[0])>mEpsilonR);
        SplinePoint  endpts[2];
        const SplinePoint  *ptrs[2];
        for (int i=0; i<2; ++i)
          {
            ptrs[i] = & endpts[i];
            endpts[i].mST(0,0)=inpts[i]->mS;
            assert(Matrix::Abs(t-inpts[i]->mT)<mEpsilon);
            endpts[i].mST(0,1)=inpts[i]->mT;
            if ( (inpts[i]->mNormal[0]>0) == increasingt[i] )
              {
                endpts[i].mTangent(0,0)= - inpts[i]->mNormal[1];
                endpts[i].mTangent(0,1)= + inpts[i]->mNormal[0];
                endpts[i].mCurvature = - inpts[i]->mCurvature;
              }
            else
              {
                endpts[i].mTangent(0,0)= + inpts[i]->mNormal[1];
                endpts[i].mTangent(0,1)= - inpts[i]->mNormal[0];
                endpts[i].mCurvature = + inpts[i]->mCurvature;
              }

            // normalize tangent length
            ScalarType tanlen = Matrix::TwoNorm(endpts[i].mTangent);
            Matrix::Scale(endpts[i].mTangent, 1/tanlen);
          }
        SimpleJoin(ptrs, controlpts);
      }

#else
    
    void Cubic2dContourHelper 
    ::FindControlPoints(SplinePoint *inpts[3], STLocation controlpts[2])
      {
#error FindControlPoints not implemented for non-curvature case
      }

    
    void Cubic2dContourHelper 
    ::OptimizeSplinePath(int rootnum, int totalnumroots, 
                          SplinePoint  pts[3],
                          std::vector<STLocation> &results)
      {
        // implement this
        // like the curvature version above, but uses three points at a
        // time, as required by FindControlPoints above
#error OptimizeSplinePath not implemented for non-curvature case
      }

#endif

    
    int Cubic2dContourHelper 
    ::RotateSide(int rotatecounterclockwise, int torotate)
      {
        assert(torotate!=-1);
        if (rotatecounterclockwise)
          {
            int remainder = torotate % 3;
            torotate += rotatecounterclockwise;
            if (remainder+rotatecounterclockwise>=3)
              torotate-=3;
          }
        return torotate;
      }

    void Cubic2dContourHelper 
    ::TransitionToSplinePoint(const TransitionPoint &tp, SplinePoint &sp)
      {
        ScalarType n[2];
        n[0]=tp.mNormal[0];
        n[1]=tp.mNormal[1];
        ScalarType tangentfix=1/sqrt(n[0]*n[0]+n[1]*n[1]);
        n[0]*=tangentfix;
        n[1]*=tangentfix;
        bool left;
        if (tp.mTransition==EndAtEdge)
          // ensure leaving edge
          switch (tp.mSide)
            {
              case 0:
                assert(0);
                break;
              case 1:
                left = (n[0] - n[1] > mEpsilon) ||
                       (n[0] - n[1] > -mEpsilon &&
                        n[0] + n[1] > mEpsilon);
                break;
              case 2:
                left = (n[1]>mEpsilon) ||
                       (n[1]>-mEpsilon && 
                        n[0]>=0) ;
                break;
              case 3:
                assert(0);
                break;
              case 4:
                assert(0);
                break;
              case 5:
                left = (n[0]>0);
                break;
            }
        else if (tp.mTransition==StartAtEdge)
          // ensure entering edge
          switch (tp.mSide)
            {
              case 0:
                left = (n[0]>0);
                break;
              case 4:
              case 1:
                left = (n[1]>0);
                break;
              case 2:
              case 3:
                left = (n[1] < n[0]);
                break;
              case 5:
                assert(0);
                break;
            }
        else
          // ensure t component of tangent is positive
          left=(n[0]>0);

        if (left)
          {
            sp.mTangent(0,0) = -n[1];
            sp.mTangent(0,1) = n[0];
            // normal to right of tangent, 
            // so flip sign of curvature
            sp.mCurvature = -tp.mCurvature;
          }
        else
          {
            sp.mTangent(0,0) = n[1];
            sp.mTangent(0,1) = -n[0];
            sp.mCurvature = tp.mCurvature;
          }
      }

    void Cubic2dContourHelper 
    ::UnknownToKnown(const SplinePoint &u, const SplinePoint &k, 
                     STLocation &out1, STLocation &out2)
      { // a quadratic curve is completely determined by
        // the location of the two endpoints, and the tangent and
        // curvature at one of the endpoints
        // Know: u.mST  (c_0)
        //       k.mST  (c_2)
        //       k.mTangent
        //       k.mCurvature
        // c''=2(c_2-c_0)-/+4s*tangent
        // l = |c''-Dot(c'',tangent)tangent|
        // Note: l does not depend on s
        // s = l/2|curvature|
        // c_1=c_2+/-s*tangent
        STLocation c2mc0=k.mST;
        Matrix::Subtract(c2mc0, c2mc0, u.mST);
        STLocation cpp;
        Matrix::EqScale(cpp, 2, c2mc0);
        ScalarType dot=Matrix::Dot(cpp, k.mTangent);
        Matrix::AddScaled(cpp, -dot, k.mTangent);
        ScalarType cppl = Matrix::TwoNorm(cpp);
        STLocation c1 = k.mST;
        ScalarType s = cppl/(2*Matrix::Abs(k.mCurvature));
        assert(s>0);
        // don't let s get too large
        ScalarType l=Matrix::NormSq(c2mc0);
        dot=Matrix::Abs(dot);
        if (s*dot/2 > l)
          s=l*2/dot;

        Matrix::AddScaled(c1, -s, k.mTangent);
        // to raise quadratic u.mST, c1, k.mST to a cubic, use:
        // u.mST, (2*c1+u.mST)/3, (2*c1+k.mST)/3, k.mST
        Matrix::EqScale(out2, 2./3, c1);
        out1=out2;
        Matrix::AddScaled(out1, 1./3, u.mST);
        Matrix::AddScaled(out2, 1./3, k.mST);
      }
  }
