#ifndef __MATRIX_OPS_H
#define __MATRIX_OPS_H

#include <math.h>
#include <assert.h>

namespace Matrix {

// Not nearly as clever as Blitz++ (or valarrays), 
// but will at least allow me to vary matrix 
// representation and numeric type

// Matrix types need:
  // typedef ... Number;
  // default and copy constructor
  // const methods: int Width(), int Height()
  // const and non-const operator() taking two parameters
  //  (non-const version should return a reference)
  // typedef ... RowType, ColumnType, and TransposeType;
  // const methods: void InitializeRow(RowType &c),
  //                void InitializeColumn(ColumnType &),
  //                void InitializeTranspose(TransposeType &c)

// Matrix Operations ---------------------------------------

// Begin re-invention of the wheel

template <class N>
inline N Sgn(N t)
  {
    return (t<0) ? -1 : 1;
  }

template <class N>
inline N Abs(N t)
  {
    return (t<0) ? -t : t;
  }

template <class N>
inline N Max(N t1, N t2)
  {
    return (t1<t2) ? t2 : t1;
  }

template <class N>
inline N Min(N t1, N t2)
  {
    return (t1<t2) ? t1 : t2;
  }

template <class M1, class M2>
M1::Number Dot(const M1 & m1, const M2 & m2)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Height() == m2.Height());
    M1::Number sum = 0;
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        sum += m1(r,c)*m2(r,c);
    return sum;
  }

template <class M>
M::Number NormSq(const M & m)
  {
    M::Number sumsq = 0;
    for (int r=0; r<m.Height(); ++r)
      for (int c=0; c<m.Width(); ++c)
        {
          M::Number t=m(r,c);
          sumsq += t*t;
        }
    return sumsq;
  }

template <class M>
M::Number TwoNorm(const M & m)
  {
    return sqrt(NormSq(m));
  }

template <class M>
void Scale(M & m, M::Number scale)
  {
    for (int r=0; r<m.Height(); ++r)
      for (int c=0; c<m.Width(); ++c)
        m(r,c)*=scale;
  }

template <class M>
bool IsSquare(const M & m)
  {
    return m.Width()==m.Height();
  }

// m1 = m2 + m3
template <class M1, class M2, class M3>
void Add(M1 &m1, const M2 & m2, const M3 & m3)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Width() == m3.Width());
    assert(m1.Height() == m2.Height());
    assert(m1.Height() == m3.Height());
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c) = m2(r,c)+m3(r,c); 
  }

// m1 = m2 - m3
template <class M1, class M2, class M3>
void Subtract(M1 &m1, const M2 & m2, const M3 & m3)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Width() == m3.Width());
    assert(m1.Height() == m2.Height());
    assert(m1.Height() == m3.Height());
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c) = m2(r,c)-m3(r,c); 
  }

// m1 = scale * m2
template <class M1, class M2>
void EqScale(M1 & m1, M2::Number scale, const M2 & m2)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Height() == m2.Height());

    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c)=scale*m2(r,c);
  }

// m1 += scale * m2
template <class M1, class M2>
void AddScaled(M1 & m1, M2::Number scale, const M2 & m2)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Height() == m2.Height());

    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c)+=scale*m2(r,c);
  }

// m1 += scale * (m2-m3)
template <class M1, class M2, class M3>
void AddScaledDiff(M1 & m1, M1::Number scale, const M2 & m2, const M3 & m3)
  {
    assert(m1.Width() == m2.Width());
    assert(m2.Width() == m3.Width());
    assert(m1.Height() == m2.Height());
    assert(m2.Height() == m3.Height());

    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c)+=scale*(m2(r,c)-m3(r,c));
  }

// m1 = transpose(m2)
template <class M1, class M2>
void Transpose(M1 &m1, const M2 & m2)
  {
    assert(m1.Width() == m2.Height());
    assert(m1.Height() == m2.Width());
    assert(((void *)&m1) != ((void *)&m2));
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c) = m2(c,r);
  }

// m1 = transpose(m1)
template <class M>
void TransposeInPlace(M &m)
  {
    assert(IsSquare(m));
    for (int r=0; r<m.Height(); ++r)
      for (int c=r+1; c<m.Width(); ++c)
        std::swap(m(r,c), m(c,r));
  }

template <class M1, class M2>
M2::Number Length(const M1 & m1, const M2 &m2)
  {
    assert(m1.Width() == m2.Width());
    assert(m1.Height() == m2.Height());

    M2::Number sumsq = 0;
    for (int r=0; r<m2.Height(); ++r)
      for (int c=0; c<m2.Width(); ++c)
        {
          M2::Number t=m2(r,c)-m1(r,c);
          sumsq += t*t;
        }
    return sqrt(sumsq);
  }



// m1 = m2 * m3
template <class M1, class M2, class M3>
void Mult(M1 &m1, const M2 & m2, const M3 & m3)
  {
    assert(m1.Height() == m2.Height());
    assert(m2.Width() == m3.Height());
    assert(m1.Width() == m3.Width());
    assert(((void *)&m1) != ((void *)&m2));
    assert(((void *)&m1) != ((void *)&m3));
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        {
          M1::Number temp=0;
          for (int i=0; i<m2.Width(); ++i)
            temp += m2(r,i)*m3(i,c);
          m1(r,c)=temp;
        }
  }

// m1 = m2 * m1
template <class M1, class M2>
void PreMult(M1 &m1, const M2 & m2)
  {
    assert(m1.Height() == m2.Height());
    assert(IsSquare(m2));
    assert(((void *)&m1) != ((void *)&m2));
    M1::ColumnType column;
    m1.InitializeColumn(column);
    for (int c=0; c<m1.Width(); ++c)
      {
        int r;
        for (r=0; r<m1.Height(); ++r)
          column(r,0)=m1(r,c);
        for (r=0; r<m1.Height(); ++r)
          {
            M1::Number temp=0;
            for (int i=0; i<m1.Height(); ++i)
              temp += m2(r,i)*column(i,0);
            m1(r,c)=temp;
          }
      }
  }

// m1 = m1 * m2
template <class M1, class M2>
void PostMult(M1 &m1, const M2 & m2)
  {
    assert(m1.Width() == m2.Width());
    assert(IsSquare(m2));
    assert(((void *)&m1) != ((void *)&m2));
    M1::RowType row;
    m1.InitializeRow(row);
    for (int r=0; r<m1.Height(); ++r)
      {
        int c;
        for (c=0; c<m1.Width(); ++c)
          row(0,c)=m1(r,c);
        for (c=0; c<m1.Width(); ++c)
          {
            M1::Number temp=0;
            for (int i=0; i<m2.Width(); ++i)
              temp += m2(i,c)*row(0,i);
            m1(r,c)=temp;
          }
      }
  }

// Solve R x = u with R upper triangular and square
template <class M1, class M2, class M3>
void SolveUpperTri(M1 &x, const M2 & R, const M3 & u)
  {
    assert(x.Width()==1);
    assert(u.Width()==1);
    assert(R.Width()==x.Height());
    assert(R.Height()==u.Height());
    assert(x.Height()==u.Height());
    for (int i=x.Height()-1; i>=0; --i)
      {
        M3::Number y = u(i,0);
        for (int j=x.Height()-1; j>i; --j)
          y -= R(i,j)*x(j,0);
        x(i,0) = y / R(i,i);  // should I check for division by zero?
      }
  }

// set m to the identity matrix
template <class M>
void SetToId(M & m)
  {
    assert(IsSquare(m));
    for (int r=0; r<m.Height(); ++r)
      for (int c=0; c<m.Width(); ++c)
        m(r,c)=(r==c);  // decided to go this way for cache reasons...
  }
// specialize for sparse matrices...

// m1 = m2
template <class M1, class M2>
void Set(M1 & m1, M2 & m2)
  {
    assert(m1.Width()==m2.Width());
    assert(m1.Height()==m2.Height());
    for (int r=0; r<m1.Height(); ++r)
      for (int c=0; c<m1.Width(); ++c)
        m1(r,c)=m2(r,c);
  }



template <class M, class ostream>
void Print(M & m, ostream & outto)
  {
    for (int r=0; r<m.Height(); ++r)
      {
        outto << "[ ";
        for (int c=0; c<m.Width(); ++c)
          {
            outto << m(r,c) << "\t";
          }
        outto << "]\n";
      }
  }

}
#endif