#ifndef __LINEAR_MESH_H
#define __LINEAR_MESH_H

#include "matrix.h"
#include "vert.h"
#include <vector>
#include <math.h> // for ceil/floor
#include <stdio.h> // for printf

template<class Traits_T>
class LinearMesh
  {
public:
    typedef Traits_T TraitsType;
    typedef LinearMesh<TraitsType> thistype;
    typedef typename TraitsType::DomainType PositionType;
    typedef TraitsType::VertexInfo VertexType;
    typedef int ElementType; 

protected:
    class LinearElement
      {
        friend class LinearMesh<TraitsType>;
        int mVerts[2];
        int mCenter;
      };

    typedef std::vector<VertexType> VertPool;
    VertPool mVerts;

    typedef std::vector<LinearElement> ElementPool;
    ElementPool mElements;

    typedef std::vector<ElementType> ListType;  //list, vector, or s_list?
    typedef ListType::iterator    IterType;
    ListType mWorkingSet;
    ListType mResultSet;

    PositionType mMinDom, mMaxDom;
    

    bool InWorkingOrResult(ElementType element)
      {
        {
          for (IterType i=mWorkingSet.begin(); i!=mWorkingSet.end(); ++i)
            if ((*i)==element)
              return true;
        }
        {
          for (IterType i=mResultSet.begin(); i!=mResultSet.end(); ++i)
            if ((*i)==element)
              return true;
        }
        return false;
      }
public:
        // Initialize(...)
        // Creates an initial mesh and inserts it into the working set
    void Initialize(PositionType minDom, PositionType maxDom)
      {
        //assert(mVerts.empty());
        //assert(mElements.empty());
        mVerts.clear();
        mElements.clear();
        assert(mWorkingSet.empty());
        assert(mResultSet.empty());

        mMinDom=minDom;
        mMaxDom=maxDom;
        mVerts.push_back(VertexType(minDom));
        mVerts.push_back(VertexType(maxDom));
        mElements.push_back();
        mElements.back().mVerts[0]=0;
        mElements.back().mVerts[1]=1;
        mElements.back().mCenter = -1;
        mWorkingSet.push_back(0);
      }

      // Element management functions ------------------------------------
    static ElementType InvalidElement()
      { return -1; }

    bool ElementValid(ElementType element)
      {
        return (element>=0 && element<mElements.size());
      }
protected:
    bool VertexValid(int vert)
      {
        return (vert>=0 && vert<mVerts.size());
      }

public:
    static int NumCorners(ElementType)
      { return 2; }

    int Neighbor(ElementType, int)
      { return 0; /* not used */ }

    VertexType & Corner(ElementType element, int corner)
      {
        assert(corner>=0);
        assert(corner<2);
        assert(ElementValid(element));
        return mVerts[mElements[element].mVerts[corner]];
      }

    ScalarType Diameter(ElementType element)
      {
        assert(ElementValid(element));
        return Matrix::Length(Corner(element,0).mPos, Corner(element,1).mPos);
      }

    ScalarType Radius(ElementType element)
      {
        return Diameter(element)*.5;
      }

    bool IsCoveredByCorners(ElementType element, ScalarType elementradius)
      {
        assert(ElementValid(element));
        assert(elementradius>0);
        return (2*elementradius <= Corner(element, 0).mSafeRadius + 
                                   Corner(element, 1).mSafeRadius);
      }

    void AllocateCenter(ElementType element) 
      { 
        assert(ElementValid(element));
        if (mElements[element].mCenter == -1)
          {
            PositionType centerpos;
            Matrix::Add(centerpos, Corner(element, 0).mPos, Corner(element,1).mPos);
            Matrix::Scale(centerpos, 0.5);
            mVerts.push_back(VertexType(centerpos));
            mElements[element].mCenter=mVerts.size()-1;
          }
        assert(VertexValid(mElements[element].mCenter));
      }
    
    VertexType & Center(ElementType element)
      {
        assert(ElementValid(element));
        assert(VertexValid(mElements[element].mCenter));
        return mVerts[mElements[element].mCenter];
      }

    ScalarType RadiusForCenterToCover(ElementType element, ScalarType elementradius)
      {
        assert(ElementValid(element));
        assert(elementradius>0);
        return elementradius-Matrix::Min(Corner(element, 0).mSafeRadius, 
                                         Corner(element, 1).mSafeRadius);
      }

      // --------------------------------------------------------------

        // PopWorking(...)
        // removes a element from the working set and returns it
        // returns InvalidElement() if working set is empty
        // result should be then either: refined, discarded, or 
        //        added to the results
    ElementType PopWorking()
      {
        if (mWorkingSet.empty())
          return InvalidElement();
        else
          {
            ElementType ret=mWorkingSet.back();
            mWorkingSet.pop_back();
            return ret;
          }
      }

        // Refine(...)
        // refine must not be in the working set or result set
    void Refine(ElementType refine)
      {
        assert(ElementValid(refine));
        assert(!InWorkingOrResult(refine));
        AllocateCenter(refine);
        mElements.push_back();
        mElements.back().mVerts[0]=mElements[refine].mCenter;
        mElements.back().mVerts[1]=mElements[refine].mVerts[1];
        mElements.back().mCenter=InvalidElement();
        mElements[refine].mVerts[1]=mElements[refine].mCenter;
        mElements[refine].mCenter=InvalidElement();
        mWorkingSet.push_back(mElements.size()-1);
        mWorkingSet.push_back(refine);
      }

        // Currently does nothing
        // could potentially reduce memory usage here
        // if I feel like visualizing, probably want to keep this data around
    void DiscardElement(ElementType element)
      {
      }

        // PushResult(...)
        // adds a element to the result set
        // must not (already) be part of the result or working set
    void PushResult(ElementType result)
      {
        assert(ElementValid(result));
        assert(!InWorkingOrResult(result));
        mResultSet.push_back(result);
      }

        // PopWorking(...)
        // removes a element from the result set and returns it
        // returns InvalidElement() if result set is empty
    ElementType PopResult()
      {
        if (mResultSet.empty())
          return InvalidElement();
        else
          {
            ElementType ret=mResultSet.back();
            mResultSet.pop_back();
            return ret;
          }
      }

    int ResultSetSize()
      {
        return mResultSet.size();
      }

    void Visualize()
      {
        //  -- working set
        //  .. discarded
        //  == result set
        //   ^ vertex > 0
        //   v vertex < 0
        //   0 vertex = 0
        //   ? vertex not evaluated
        // vertex postion mMinDom -> 0
        // vertex postion mMaxDom -> screenwidth-1
        const int screenwidth = 79;
        char display[screenwidth+1];
        for (int i=0; i<screenwidth; ++i)
          display[i]='.';
        display[screenwidth]=0;

        IterType cur;
        for (cur = mWorkingSet.begin(); cur != mWorkingSet.end(); ++cur)
          {
            ElementType element = *cur;
            VertexType *verts[2];
            verts[0]=& mVerts[mElements[element].mVerts[0]];
            verts[1]=& mVerts[mElements[element].mVerts[1]];
            int pos[2];
            pos[0]=ceil ((verts[0]->mPos(0,0)-mMinDom(0,0))*(screenwidth-1)/(mMaxDom(0,0)-mMinDom(0,0)));
            pos[1]=floor((verts[1]->mPos(0,0)-mMinDom(0,0))*(screenwidth-1)/(mMaxDom(0,0)-mMinDom(0,0)));
            for (; pos[0]<=pos[1]; ++pos[0])
              display[pos[0]]='-';
          }

        for (cur = mResultSet.begin(); cur != mResultSet.end(); ++cur)
          {
            ElementType element = *cur;
            VertexType *verts[2];
            verts[0]=& mVerts[mElements[element].mVerts[0]];
            verts[1]=& mVerts[mElements[element].mVerts[1]];
            int pos[2];
            pos[0]=ceil ((verts[0]->mPos(0,0)-mMinDom(0,0))*(screenwidth-1)/(mMaxDom(0,0)-mMinDom(0,0)));
            pos[1]=floor((verts[1]->mPos(0,0)-mMinDom(0,0))*(screenwidth-1)/(mMaxDom(0,0)-mMinDom(0,0)));
            for (; pos[0]<=pos[1]; ++pos[0])
              display[pos[0]]='=';
          }

        VertPool::iterator v;
        for (v=mVerts.begin(); v!=mVerts.end(); ++v)
          {
            int p = floor(((*v).mPos(0,0)-mMinDom(0,0))*(screenwidth-1)/(mMaxDom(0,0)-mMinDom(0,0)) +.5);
            if ((*v).mEvaluated)
              if ((*v).mFunctionValue>0)
                display[p]='^';
              else if ((*v).mFunctionValue<0)
                display[p]='v';
              else
                display[p]='0';
            else
              display[p]='?';
          }
        printf("%s\n", display);
      }
  };

#endif