// included by meshbtt.h, contains implementations of functions
#ifndef __BTT_MESH_H
#error meshbtt.hpp should only be included from meshbtt.h
#endif

#include "matrix.h"
#include "vert.h"
#include <fixme.h>
#include <vector>
#include <algorithm>

/* A leaf-only split-only binary-triangle-tree mesh data structure
   a.k.a. newest-vertex-bisection mesh

    Tricky parts:
      * Elements already placed in the result set may need to be refined
        in order to maintain a conforming mesh.  Therefore elements need
        to be able to quickly tell if they are in the result set.
      * Internally the mesh contains (right-isoceles) triangles, but the
        refinement operation is normally applied to two trianles meeting
        base-to-base, ie. they form a "diamond".  To the outside world,
        it reports an element as the union of the two triangles if they
        form a diamond.  Only one member of the diamond is put into the
        working set, but both are put in the result set.
      * In order to avoid having to remove a random element from the 
        working set, the working set is ordered by the diameter of the
        elements.  PopWorking() returns the largest diameter element.
        This should elimate force-splits of elements in the working set
        (elements in the result set still may need to be force split:
        this is the only way elements in the result set are split;
        elements discarded shouldn't: they are disconnected from the 
        topology).  It should also improve the number of triangles 
        considered as part as a diamond instead of individually.
      * Currently mNumResults counts the number of triangles in the 
        result set, not the number of times PopResult() will return
        an element (since it returns both triangles and diamonds).

    Assumptions:
      * The following functions assume elements are right-isoceles:
          Diameter(...)
          IsCoveredByCorners(...)
          RadiusForCenterToCover(...), along with its helper function:
          RadiusToCoverCorner(...)
      * PopWorking implicitly assumes that elements only force split
        elements with a larger diameter.  That is the whole point of
        maintaining a priority queue for the working set ordered by
        diameter.

    Future:
      * If PopResult() is too slow, decent asymptotic behavoir can be
        achieved using a set or a hash_set to store which elements are
        in the result set.
 */

template<class Traits_T>
void BTTMesh<Traits_T>::PushToWorkingSet(ElementType elt)
  {
    mWorkingSet.push_back(elt);
    std::push_heap(mWorkingSet.begin(), mWorkingSet.end(), mCompare);
  }

template<class Traits_T>
BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::PopFromWorkingSet()
  {
    assert(! mWorkingSet.empty());
    std::pop_heap(mWorkingSet.begin(), mWorkingSet.end(), mCompare);
    ElementType ret=mWorkingSet.back();
    mWorkingSet.pop_back();
    return ret;
  }

template<class Traits_T>
inline BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::InvalidElement()
  { return -1; }

template<class Traits_T>
inline bool BTTMesh<Traits_T>::ElementValid(ElementType element)
  {
    return (element>=0 && element<mElements.size());
  }

template<class Traits_T>
inline bool BTTMesh<Traits_T>::IsSameElement(ElementType elt1, ElementType elt2)
  {
    if (elt1==elt2)
      return true;
    if (! ElementValid(elt1) || ! ElementValid(elt2))
      return false;
    // can be same element if together form a diamond
    return (mElements[elt1].mNeighbors[Base]==elt2 &&
            mElements[elt2].mNeighbors[Base]==elt1 &&
            InResult(elt1)==InResult(elt2));
  }

template<class Traits_T>
inline bool BTTMesh<Traits_T>::InWorking(ElementType element)
  {
    {
      for (IterType i=mWorkingSet.begin(); i!=mWorkingSet.end(); ++i)
        if (IsSameElement(*i, element))
          return true;
    }
    return false;
  }

template<class Traits_T>
inline bool BTTMesh<Traits_T>::InResult(ElementType element)
  {
    assert(ElementValid(element));
    return mElements[element].mNextResult!=NotResult;
  }

template<class Traits_T>
inline bool BTTMesh<Traits_T>::InWorkingOrResult(ElementType element)
  {
    return InWorking(element) || InResult(element);
  }
  
template<class Traits_T>
BTTMesh<Traits_T>::BTTMesh()
    // Initialize the result set to empty
  : mNumResults(0), mFirstResult(InvalidElement()), mCompare(this)
  { }

    // Initialize ...(...)
    // Creates an initial mesh and inserts it into the working set

    // forms diamond a, b, c, a+c-b
template<class Traits_T>
void BTTMesh<Traits_T>::InitializeDiamond(PositionType a, PositionType b, PositionType c)
  {
    PositionType d;
    Matrix::Add(d, a, c);
    Matrix::Subtract(d, d, b);
    mVerts.push_back(VertexType(a));
    mVerts.push_back(VertexType(b));
    mVerts.push_back(VertexType(c));
    mVerts.push_back(VertexType(d));
    mElements.push_back();
    mElements.back().mVerts[Left] =mVerts.size()-4;
    mElements.back().mVerts[Right]=mVerts.size()-2;
    mElements.back().mVerts[Base] =mVerts.size()-3;
    mElements.back().mCenter = -1;
    mElements.back().mNeighbors[Left] = -1;
    mElements.back().mNeighbors[Right]= -1;
    mElements.back().mNeighbors[Base] = mElements.size();
    mElements.back().mNextResult = NotResult;
    mElements.push_back();
    mElements.back().mVerts[Left] =mVerts.size()-2;
    mElements.back().mVerts[Right]=mVerts.size()-4;
    mElements.back().mVerts[Base] =mVerts.size()-1;
    mElements.back().mCenter = -1;
    mElements.back().mNeighbors[Left] = -1;
    mElements.back().mNeighbors[Right]= -1;
    mElements.back().mNeighbors[Base] = mElements.size()-2;
    mElements.back().mNextResult = NotResult;
    PushToWorkingSet(mElements.size()-1);
    mInitial[0]=mVerts.size()-4;
    mInitial[1]=mVerts.size()-3;
    mInitial[2]=mVerts.size()-2;
    mInitial[3]=mVerts.size()-1;
  }

template<class Traits_T>
void BTTMesh<Traits_T>::InitializeTriangle(PositionType l, PositionType b, PositionType r)
  {
    mVerts.push_back(VertexType(l));
    mVerts.push_back(VertexType(r));
    mVerts.push_back(VertexType(b));
    mElements.push_back();
    mElements.back().mVerts[Left] =mVerts.size()-3;
    mElements.back().mVerts[Right]=mVerts.size()-2;
    mElements.back().mVerts[Base] =mVerts.size()-1;
    mElements.back().mCenter = -1;
    mElements.back().mNeighbors[Left] = -1;
    mElements.back().mNeighbors[Right]= -1;
    mElements.back().mNeighbors[Base] = -1;
    mElements.back().mNextResult = NotResult;
    PushToWorkingSet(mElements.size()-1);
    mInitial[0] = mVerts.size()-3;
    mInitial[1] = mVerts.size()-2;
    mInitial[2] = mVerts.size()-1;
    mInitial[3] = -1;
  }

  // Element management functions ------------------------------------
template<class Traits_T>
inline bool BTTMesh<Traits_T>::VertexValid(int vert)
  {
    return (vert>=0 && vert<mVerts.size());
  }

     // Examine left neighbor, reassign its pointer to new left child
     // and if a diamond has been formed, copy center vertex, otherwise
     // need to add this new element to working set
template<class Traits_T>
void BTTMesh<Traits_T>::DiamondFinder(ElementType elt, ElementType child, ElementType neighbor)
  {
    if (neighbor!=InvalidElement())
      {
        assert(ElementValid(neighbor));
        if (mElements[neighbor].mNeighbors[Base] == elt)
          {
            mElements[child].mCenter = mElements[neighbor].mCenter;
            mElements[neighbor].mNeighbors[Base] = child;
            // neighbor already in working set if not in result set, 
            // this element just completes the diamond
            if (InResult(neighbor) && !InResult(elt))
              PushToWorkingSet(child);
          }
        else 
          {
            mElements[child].mCenter = -1;
            if (mElements[neighbor].mNeighbors[Right] == elt)
              mElements[neighbor].mNeighbors[Right] = child;
            else 
              {
                assert(mElements[neighbor].mNeighbors[Left] == elt);
                mElements[neighbor].mNeighbors[Left] = child;
              }
            if (! InResult(elt))
              PushToWorkingSet(child);
          }
      }
    else
      {
        mElements[child].mCenter = -1;
        if (! InResult(elt))
          PushToWorkingSet(child);
      }
  }

template<class Traits_T>
BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::SplitHelper(ElementType elt)
  {
    assert(VertexValid(mElements[elt].mCenter));
    ElementType newchild = mElements.size();
    mElements.push_back();
    BTTElement parent = mElements[elt];
    // Left Child
    mElements[elt].mVerts[Left] =parent.mVerts[Base];
    mElements[elt].mVerts[Right]=parent.mVerts[Left];
    mElements[elt].mVerts[Base] =parent.mCenter;
    mElements[elt].mNeighbors[Left]=newchild;
    mElements[elt].mNeighbors[Base]=parent.mNeighbors[Left];
    DiamondFinder(elt, elt, parent.mNeighbors[Left]);

    // Right Child
    mElements.back().mVerts[Left] =parent.mVerts[Right];
    mElements.back().mVerts[Right]=parent.mVerts[Base];
    mElements.back().mVerts[Base] =parent.mCenter;
    mElements.back().mNeighbors[Right]=elt;
    mElements.back().mNeighbors[Base] =parent.mNeighbors[Right];
    mElements.back().mNextResult = NotResult;
    DiamondFinder(elt, newchild, parent.mNeighbors[Right]);

    if (InResult(elt))
      {
        ++mNumResults;
        ElementType rightbaseneighbor = mElements.back().mNeighbors[Base];
        if (rightbaseneighbor!=InvalidElement() &&
            InResult(rightbaseneighbor) &&
            mElements[rightbaseneighbor].mNeighbors[Base]==newchild)
          {
            mElements.back().mNextResult = mElements[rightbaseneighbor].mNextResult;
            mElements[rightbaseneighbor].mNextResult = newchild;
          }
        else
          {
            mElements.back().mNextResult = mElements[elt].mNextResult;
            mElements[elt].mNextResult = newchild;
          }
      }

    return newchild;
  }

    // Refine(...)
    // refine must not be in the working set or result set
template<class Traits_T>
void BTTMesh<Traits_T>::Refine(ElementType elt)
  {
    assert(ElementValid(elt));
    assert(!InWorking(elt));
    // WARNING: could be part of result set due to force-split!

    ElementType baseneighbor = mElements[elt].mNeighbors[Base];
    if (baseneighbor != InvalidElement())
      {
        assert(ElementValid(baseneighbor));
        if (mElements[baseneighbor].mNeighbors[Base]!=elt)
          {
            assert(InResult(baseneighbor)); // so not in working set
            Refine(baseneighbor);
            baseneighbor = mElements[elt].mNeighbors[Base];
          }
      }

    AllocateCenter(elt);

    if (baseneighbor != InvalidElement())
      {
        assert(ElementValid(baseneighbor));
        ElementType newchild = SplitHelper(elt);
        ElementType newbaseneighbor=SplitHelper(baseneighbor);
        // glue together new children
        // ideally elt.right <-> newbaseneighbor.left and
        //         newchild.left <-> baseneighbor.right
        assert(mElements[elt]            .mVerts[Right] == 
               mElements[newbaseneighbor].mVerts[Left]);
        assert(mElements[newchild]       .mVerts[Left]  == 
               mElements[baseneighbor]   .mVerts[Right]);
        mElements[elt]            .mNeighbors[Right]=newbaseneighbor;
        mElements[newchild]       .mNeighbors[Left] =baseneighbor;
        mElements[newbaseneighbor].mNeighbors[Left] =elt;
        mElements[baseneighbor]   .mNeighbors[Right]=newchild;
      }
    else
      { // no neighbors! 
        ElementType newchild = SplitHelper(elt);
        mElements[elt]     .mNeighbors[Right]=InvalidElement();
        mElements[newchild].mNeighbors[Left] =InvalidElement();
      }
  }

template<class Traits_T>
inline int BTTMesh<Traits_T>::NumCorners(ElementType elt)
  {
    assert(ElementValid(elt));
    ElementType baseneighbor = mElements[elt].mNeighbors[Base];
    if (baseneighbor != InvalidElement() &&
        mElements[baseneighbor].mNeighbors[Base]==elt &&
        InResult(elt)==InResult(baseneighbor))
      // diamond, has 4 corners
      return 4;
    else
      return 3;
  }

template<class Traits_T>
inline BTTMesh<Traits_T>::VertexType & BTTMesh<Traits_T>
::Corner(ElementType element, int corner)
  {
    assert(ElementValid(element));
    assert(corner>=0);
    assert(corner<NumCorners(element));
    switch (corner)
      {
        case 0:
          return mVerts[mElements[element].mVerts[Left]];
        case 1:
          return mVerts[mElements[element].mVerts[Base]];
        case 2:
          return mVerts[mElements[element].mVerts[Right]];
        case 3:
          return mVerts[mElements[mElements[element].mNeighbors[Base]].mVerts[Base]];
      }
    assert(0 && "Should not get here");
    // Just to supress warnings
    return mVerts[0];
  }

template<class Traits_T>
BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::Name(ElementType elt)
  {
    if (elt != InvalidElement())
      {
        assert(ElementValid(elt));
        if (NumCorners(elt)==4)
          {
            ElementType baseneighbor=mElements[elt].mNeighbors[Base];
            if (baseneighbor < elt)
              return baseneighbor;
          }
      }
    return elt;
  }


template<class Traits_T>
int BTTMesh<Traits_T>
::Neighbor(ElementType element, int which)
  {
    assert(ElementValid(element));
    assert(which>=0);
    assert(which<NumCorners(element));
    switch (which)
      {
        case 0:
          return Name(mElements[element].mNeighbors[Left]);
        case 1:
          return Name(mElements[element].mNeighbors[Right]);
        case 2:
          if (NumCorners(element)==4)
            return Name(mElements[mElements[element].mNeighbors[Base]].mNeighbors[Left]);
          else
            return Name(mElements[element].mNeighbors[Base]);
        case 3:
          return Name(mElements[mElements[element].mNeighbors[Base]].mNeighbors[Right]);
      }
    assert(0 && "Should not get here");
    // Just to supress warnings
    return -1;
  }

template<class Traits_T>
inline ScalarType BTTMesh<Traits_T>
::Diameter(ElementType element)
  {
    assert(ElementValid(element));
    return Matrix::Length(Corner(element,0).mPos, Corner(element,2).mPos);
  }

template<class Traits_T>
inline ScalarType BTTMesh<Traits_T>
::Radius(ElementType element)
  {
    return Diameter(element)*.5;
  }

template<class Traits_T>
bool BTTMesh<Traits_T>::IsCoveredByCorners(ElementType element, ScalarType elementradius)
  {
    assert(ElementValid(element));
    assert(elementradius>0);
    int corners=NumCorners(element);
    for (int c=0; c<corners; ++c)
      if (elementradius > Corner(element,c).mSafeRadius)
        return false;
    return true;
  }

template<class Traits_T>
void BTTMesh<Traits_T>::AllocateCenter(ElementType element) 
  { 
    assert(ElementValid(element));
    if (mElements[element].mCenter == -1)
      {
        PositionType centerpos;
        Matrix::Add(centerpos, Corner(element, 0).mPos, Corner(element,2).mPos);
        Matrix::Scale(centerpos, 0.5);
        mVerts.push_back(VertexType(centerpos));
        mElements[element].mCenter=mVerts.size()-1;
        ElementType baseneighbor = mElements[element].mNeighbors[Base];
        if (baseneighbor != InvalidElement() && 
            mElements[baseneighbor].mNeighbors[Base] == element)
          { // put center vertex in other half of diamond
            assert(ElementValid(baseneighbor));
            assert(mElements[baseneighbor].mCenter==-1);
            mElements[baseneighbor].mCenter = mElements[element].mCenter;
          }
      }
    assert(VertexValid(mElements[element].mCenter));
  }

template<class Traits_T>
inline BTTMesh<Traits_T>::VertexType & BTTMesh<Traits_T>::Center(ElementType element)
  {
    assert(ElementValid(element));
    assert(VertexValid(mElements[element].mCenter));
    return mVerts[mElements[element].mCenter];
  }

template<class Traits_T>
ScalarType BTTMesh<Traits_T>
::RadiusForCenterToCover(ElementType element, ScalarType elementradius)
  {
    assert(ElementValid(element));
    assert(elementradius>0);
    int numcorners = NumCorners(element);
    ScalarType ret = 0;
    for (int c=0; c<numcorners; ++c)
      {
        ScalarType compare=RadiusToCoverCorner(Corner(element, c).mSafeRadius, elementradius);
        if (compare>ret)
          ret=compare;
      }
    return ret;
  }

#define SQRT12 (1.0/1.4142135623730950488016887242097)
template<class Traits_T>
ScalarType BTTMesh<Traits_T>
::RadiusToCoverCorner(ScalarType saferadius, ScalarType elementradius)
  {
    if (saferadius > elementradius*SQRT12)
      {
        if (saferadius >= elementradius)
          return 0;
        elementradius *= SQRT12;
        return elementradius - sqrt(saferadius*saferadius - elementradius*elementradius);
      }
    else
      {
        // see if saferadius == 0 is really common
        saferadius *= SQRT12;
        ScalarType x = elementradius - saferadius;
        return sqrt(x*x+saferadius*saferadius);
      }
  }
#undef SQRT12
  // --------------------------------------------------------------

    // PopWorking(...)
    // removes a element from the working set and returns it
    // returns InvalidElement() if working set is empty
    // result should be then either: refined, discarded, or 
    //        added to the results
template<class Traits_T>
inline BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::PopWorking()
  {
    if (mWorkingSet.empty())
      return InvalidElement();
    else
      return PopFromWorkingSet();
  }

    // Currently disconnects element from adjacent elements
    // could potentially reduce memory usage here
    // if I feel like visualizing, probably want to keep this data around
template<class Traits_T>
void BTTMesh<Traits_T>::DiscardElement(ElementType element)
  {
    DisconnectFromElement(element, Left);
    DisconnectFromElement(element, Right);
    ElementType baseneighbor = mElements[element].mNeighbors[Base];
    if (baseneighbor!=InvalidElement())
      {
        assert(ElementValid(baseneighbor));
        // see NumCorners(...) test
        if (mElements[baseneighbor].mNeighbors[Base] == element &&
            InResult(element)==InResult(baseneighbor))
          {
            DisconnectFromElement(baseneighbor, Left);
            DisconnectFromElement(baseneighbor, Right);
          }
        else
          {
            DisconnectFromElement(element, Base);
          }
      }
  }

template<class Traits_T>
void BTTMesh<Traits_T>::DisconnectFromElement(ElementType element, int edge)
  {
    assert(ElementValid(element));
    assert(edge>=0 && edge<3);
    ElementType neighbor = mElements[element].mNeighbors[edge];
    if (neighbor != InvalidElement())
      {
        assert(ElementValid(neighbor));
        mElements[element].mNeighbors[edge] = InvalidElement();

        for (int e=0; e<3; ++e)
          if (mElements[neighbor].mNeighbors[e]==element)
            {
              mElements[neighbor].mNeighbors[e]=InvalidElement();
              return;
            }
        assert(0 && "Should not get here");
      }
  }

    // PushResult(...)
    // adds a element to the result set
    // must not (already) be part of the result or working set
template<class Traits_T>
inline void BTTMesh<Traits_T>::PushResult(ElementType result)
  {
    assert(ElementValid(result));
    assert(!InWorkingOrResult(result));
    if (NumCorners(result)==4)
      {
        ElementType baseneighbor = mElements[result].mNeighbors[Base];
        mElements[baseneighbor].mNextResult=mFirstResult;
        mFirstResult=baseneighbor;
        ++mNumResults;
      }

    mElements[result].mNextResult=mFirstResult;
    mFirstResult=result;
    ++mNumResults;
  }

    // PopWorking(...)
    // removes a element from the result set and returns it
    // returns InvalidElement() if result set is empty
template<class Traits_T>
BTTMesh<Traits_T>::ElementType BTTMesh<Traits_T>::PopResult()
  {
    if (mFirstResult==InvalidElement())
      {
        assert(mNumResults==0);
        return InvalidElement();
      }
    else
      {
        assert(ElementValid(mFirstResult));
        assert(mNumResults>0);
        ElementType ret=mFirstResult;
        mFirstResult=mElements[ret].mNextResult;
        --mNumResults;
        ElementType smaller=ret;
        // corners==4 => remove base neighbor from result set
        if (NumCorners(ret)==4)
          {
            ElementType baseneighbor = mElements[ret].mNeighbors[Base];
            assert(ElementValid(baseneighbor));
            assert(mElements[baseneighbor].mNeighbors[Base]==ret);
            ElementType *ptr = & mFirstResult;
            assert(ElementValid(*ptr));
            while (*ptr != baseneighbor)
              {
                ptr = & mElements[*ptr].mNextResult;
                assert(ElementValid(*ptr));
              }
            *ptr = mElements[baseneighbor].mNextResult;
            mElements[baseneighbor].mNextResult = NotResult;
            --mNumResults;
            if (baseneighbor<ret)
              smaller=baseneighbor;
          }
        mElements[ret].mNextResult=NotResult;
        return smaller;
      }
  }

template<class Traits_T>
inline int BTTMesh<Traits_T>::ResultSetSize()
  {
    return mNumResults;
  }

template<class Traits_T>
void BTTMesh<Traits_T>::Visualize()
  {
#pragma todo("Do BTT visualization")
  }
