#include "makeps.h"
#include <stdlib.h>
#include <stdio.h>
#include <assert.h>

namespace PS
  {
    // converts a number to a string and appends to end
    void Append(int i, string &result)
      {
        char buffer[64];
        itoa(i, buffer, 10);
        result += buffer;
      }

    void Append(double d, string &result)
      {
        char buffer[36];
        _snprintf (buffer, 33, "%g", d);
        result += buffer;
      }

    void AppendSpace(string &result)
      {
        if (result.length()<70)
          {
            result += " ";
          }
        else
          {
            int pos=result.rfind('\n');
            if (pos==string::npos || result.length()-pos >= 70)
              result += "\n    ";
            else
              result += " ";
          }
      }

    State::State()
      : mLineThickness(1.0), mColor("0 setgray"), mLineCap(square_cap),
        mLineJoin(miter_join), mMiterLimit(10.0)
      {
      }

    void State::AppendDifference(const State &cmpto, string &result) const
      {
        if (mLineThickness!=cmpto.mLineThickness)
          {
            Append(mLineThickness, result);
            result += " setlinewidth";
            AppendSpace(result);
          }
        if (mColor!=cmpto.mColor)
          {
            result += mColor;
            AppendSpace(result);
          }
        if (mLineCap!=cmpto.mLineCap)
          {
            Append(int(mLineCap), result);
            result += " setlinecap";
            AppendSpace(result);
          }
        if (mLineJoin!=cmpto.mLineJoin)
          {
            Append(int(mLineJoin), result);
            result += " setlinejoin";
            AppendSpace(result);
          }
        if (mMiterLimit!=cmpto.mMiterLimit)
          {
            Append(mMiterLimit, result);
            result += " setmiterlimit";
            AppendSpace(result);
          }
      }

    void State::SetGray(double g)
      { mColor = ""; 
        Append(g, mColor);
        mColor += " setgray";
      }

    void State::SetRGB(double r, double g, double b)
      { mColor = ""; 
        Append(r, mColor);
        mColor += " ";
        Append(g, mColor);
        mColor += " ";
        Append(b, mColor);
        mColor += " setrgbcolor";
      }

    Bounds::Bounds()
      {
        Clear();
      }

    void Bounds::Clear()
      {
        mBounds[0]= -1;
        mBounds[1]= -1;
        mBounds[2]= -2;
        mBounds[3]= -2;
      }

    bool Bounds::Empty() const
      {
        return mBounds[0]>mBounds[2];
      }

    void Bounds::Union(const Bounds &b)
      {
        if (Empty())
          *this = b;
        else if (!b.Empty())
          {
            if (b.mBounds[0]<mBounds[0])
              mBounds[0]=b.mBounds[0];
            if (b.mBounds[1]<mBounds[1])
              mBounds[1]=b.mBounds[1];
            if (b.mBounds[2]>mBounds[2])
              mBounds[2]=b.mBounds[2];
            if (b.mBounds[3]>mBounds[3])
              mBounds[3]=b.mBounds[3];
          }
      }

    void Bounds::Enlarge(double e)
      {
        assert(!Empty());
        mBounds[0]-=e;
        mBounds[1]-=e;
        mBounds[2]+=e;
        mBounds[3]+=e;
      }

    void Bounds::Add(double x, double y)
      {
        if (Empty())
          {
            mBounds[0]=mBounds[2]=x;
            mBounds[1]=mBounds[3]=y;
          }
        else
          {
            if (x<mBounds[0])
              mBounds[0]=x;
            else if (x>mBounds[2])
              mBounds[2]=x;
            if (y<mBounds[1])
              mBounds[1]=y;
            else if (y>mBounds[3])
              mBounds[3]=y;
          }
      }

    string Path::GetCommands() const
      {
        return "newpath "+mCommands;
      }

    void Path::MoveTo(double x, double y)
      {
        mBounds.Add(x, y);
        Append(x, mCommands);
        mCommands += " ";
        Append(y, mCommands);
        mCommands += " moveto";
        AppendSpace(mCommands);
      }

    void Path::LineTo(double x, double y)
      {
        mBounds.Add(x, y);
        Append(x, mCommands);
        mCommands += " ";
        Append(y, mCommands);
        mCommands += " lineto";
        AppendSpace(mCommands);
      }

    void Path::CurveTo(double pts[6])
      {
        for (int i=0; i<6; i+=2)
          {
            mBounds.Add(pts[i], pts[i+1]);
            for (int j=i; j<i+2; ++j)
              {
                Append(pts[j], mCommands);
                mCommands += " ";
              }
          }
        mCommands += "curveto";
        AppendSpace(mCommands);
      }

    void Path::Arc(double x, double y, double r, 
                   double degrees1, double degrees2)
      {
        mBounds.Add(x-r, y-r);  // conservative
        mBounds.Add(x+r, y+r);

        Append(x, mCommands);
        mCommands += " ";
        Append(y, mCommands);
        mCommands += " ";
        Append(r, mCommands);
        mCommands += " ";
        Append(degrees1, mCommands);
        mCommands += " ";
        Append(degrees2, mCommands);
        mCommands += " arc";
        AppendSpace(mCommands);
      }

    void Path::Circle(double x, double y, double r)
      {
        MoveTo(x+r, y);
        Arc(x, y, r, 0, 360);
        Close();
      }

    void Path::Close()
      {
        mCommands += "closepath";
        AppendSpace(mCommands);
      }

    void Path::Outline(const State &state)
      {
        // new path commands = 
        // save line width, cap, join, and miter limit
        // old path commands
        // set line width, cap, join, and miter limit
        // "strokepath"
        // restore line width, cap, join, and miter limit
        mCommands += "currentlinewidth "
                     "currentlinecap "
                     "currentlinejoin "
                     "currentmiterlimit";
        AppendSpace(mCommands);
        Append(state.mLineThickness, mCommands);
        mCommands += " setlinewidth";
        AppendSpace(mCommands);
        Append(int(state.mLineCap), mCommands);
        mCommands += " setlinecap";
        AppendSpace(mCommands);
        Append(int(state.mLineJoin), mCommands);
        mCommands += " setlinejoin";
        AppendSpace(mCommands);
        Append(state.mMiterLimit, mCommands);
        mCommands += " setmiterlimit";
        AppendSpace(mCommands);
        mCommands += "strokepath";
        AppendSpace(mCommands);
        mCommands += "setcurrentmiterlimit "
                     "setlinejoin "
                     "setlinecap "
                     "setlinewidth";
        AppendSpace(mCommands);

        // new path bounds = old bounds enlarged by line width / 2
        mBounds.Enlarge(state.mLineThickness/2);
      }

    void Path::Clear()
      {
        mCommands = "";
        mBounds.Clear();
      }

    Drawing::Drawing()
      {
        pathnum=0;
      }

    Path Drawing::SavePath(const Path &path)
      {
        string name("path");
        Append(pathnum, name);
        ++pathnum;

        mCommands+="/";
        mCommands+=name;
        mCommands+=" {";
        mCommands+=path.mCommands;
        mCommands+=" } bind def\n";

        Path ret;
        ret.mCommands=name;
        ret.mBounds=path.mBounds;
        return ret;
      }

    void Drawing::StrokePath(const Path &path, const State &state)
      {
        mCommands += path.GetCommands();
        state.AppendDifference(mState, mCommands);
        mState=state;
        mCommands += "stroke\n";
        Bounds bounds=path.mBounds;
        bounds.Enlarge(mState.mLineThickness/2);
        mBounds.Union(bounds);
      }

    void Drawing::FillPath(const Path &path, const State &state)
      {
        mCommands += path.GetCommands();
        state.AppendDifference(mState, mCommands);
        mState=state;
        mCommands += "fill\n";
        mBounds.Union(path.mBounds);
      }

    void Drawing::DrawText(const char *str, 
                  double x, double y,
                  double size, const char *font, 
                  const State &state) // state includes color
      { // INCOMPLETE!
        mCommands += "/";
        mCommands += font;
        mCommands += " findfont ";
        Append(size, mCommands);
        mCommands += " scalefont setfont ";
        Append(x, mCommands);
        mCommands += " ";
        Append(y, mCommands);
        mCommands += " moveto (";
        mCommands += str;
        mCommands += ") show";
        AppendSpace(mCommands);
      }

    // todo: have a common prolog (which defines a bunch 
    // of abbreviations to be used above)
    void Drawing::WritePS(string &result, const string &creator, 
                          const string &title) const
      {
        result="%!PS-Adobe-3.0\n"
               "%%BoundingBox:";
        for (int i=0; i<4; ++i)
          {
            result += " ";
            Append(int(mBounds.mBounds[i]+(i>1 ? 1 : 0)), result);
          }
        // get date so we can set CreationDate:
        result += "\n%%Creator: ";
        result += creator;
        result += "\n%%Title: ";
        result += title;
        // %%DocumentNeededResources: font Palatino-Roman Helvetica
        result += "\n%%Pages: 1\n"
                  "%%PageOrder: Ascend\n"
                  // %%DocumentNeededFonts: ...
                  // %%DocumentFonts: ...
                  // (list all fonts used in both)
                  "%%EndComments\n"
                  "%%BeginProlog\n"
                  "%%EndProlog\n"
                  "%%BeginSetup\n"
                  "%%EndSetup\n"
                  "%%Page: 1 1\n"
                  "%%BeginPageSetup\n"
                  "%%EndPageSetup\n"
                  "0.5 setflat\n";
        result += mCommands;
        result += "showpage\n"
                  "%%EOF\n";
      }

    void Drawing::WriteEPS(string &result, 
                           const string &creator, 
                           const string &title) const
      {
        result="%!PS-Adobe-3.0 EPSF-3.0\n"
               "%%BoundingBox:";
        for (int i=0; i<4; ++i)
          {
            result += " ";
            Append(int(mBounds.mBounds[i]+(i>1 ? 1 : 0)), result);
          }
        // get date so we can set CreationDate:
        result += "\n%%Creator: ";
        result += creator;
        result += "\n%%Title: ";
        result += title;
        // %%DocumentNeededFonts: ...
        // %%DocumentFonts: ...
        // (list all fonts used in both)
        result += "\n%%EndComments\n"
                  "save\n";
        Append(pathnum, result);
        result += " dict begin\n" 
                  "0.5 setflat\n";
        result += mCommands;
        result += "end\n"
                  "restore\n"
                  "showpage\n"
                  "%%EOF\n";
      }
  }