#ifndef __QR_MATRIX_DECOMPOSITION_H
#define __QR_MATRIX_DECOMPOSITION_H

#include "matrix.h"

namespace Matrix {
// Decomposes R into QR with inverse(Q)=transpose(Q)
// and R upper triangular ** overwrites R **
// Uses H as temporary space ** overwrites H **
// Follows "Numerical mathematics" by Hammerlin and Hoffman pages 64-67
// (speaking of which: mistake page 66: C_{\mu-1} is not a square matrix)
template <class MQ, class MR, class MH>
void QRDecompositionSlow(MQ & Q, MR & R, MH & H)
  { // All matrices should be distinct, square, and the same size
    assert(IsSquare(Q));
    assert(IsSquare(R));
    assert(IsSquare(H));
    assert(Q.Width()==R.Width());
    assert(Q.Width()==H.Width());
    assert(((void *)&Q) != ((void *)&R));
    assert(((void *)&Q) != ((void *)&H));
    assert(((void *)&R) != ((void *)&H));
    for (int mu=0; mu<R.Width()-1; ++mu)
      {
        // a = bottom Height-mu rows of mu-th column

        // Compute norm a
        MH::Number norm_a=0;
        int r;
        for (r=mu; r<R.Height(); ++r)
          {
            MH::Number ar = R(r, mu);
            norm_a += ar*ar;
          }
        norm_a = sqrt(norm_a);

        // Compute beta, u
          {
            MH::Number u0 = Abs(R(mu,mu))+norm_a;
            MH::Number beta = 1/(norm_a*u0);
            // u0 *= Sgn(R(mu,mu)); // replaced with:
            if (R(mu,mu)<0) u0 = -u0;
            MH::ColumnType u;
            H.InitializeColumn(u);
            u(mu,0)=u0;
            for (r=mu+1; r<R.Height(); ++r)
              u(r,0)=R(r,mu);
            // would it be better to multiply u by sqrt(beta) instead
            // of multiplying by beta afterwards?  I'll assume that
            // it causes accuracy problems and play it safe...

            // Compute H twiddle = I - beta * u * transpose(u)
            // in the lower right of H
            for (r=mu; r<R.Height(); ++r)
              {
                for (int c=mu; c<r; ++c)
                  {
                    H(c,r) = -beta*u(c,0)*u(r,0);
                    H(r,c) = H(c,r);
                  }
                MH::Number arr = u(r, 0);
                H(r,r)=1-beta*arr*arr;
              }
          }
        // pretend that we have an actual H
        // (i.e. we filled the top and left with identity matrix)

        // R = H * R
          {
            MR::ColumnType column;
            R.InitializeColumn(column);
            for (int c=0; c<R.Width(); ++c) // maybe handle c==mu specially?
              {
                for (r=mu; r<R.Height(); ++r)
                  column(r,0) = R(r, c);
                for (r=mu; r<R.Height(); ++r)
                  {
                    MR::Number temp=0;
                    for (int i=mu; i<R.Width(); ++i)
                      temp += H(r, i)*column(i,0);
                    R(r,c)=temp;
                  }
              }
          }
        // First mu+1 columns should be upper triangular
        //  std::cout << "mu = " << mu << " R =\n";
        //  Print(R, std::cout);

        // Q = Q * H
        if (mu==0)
          {
            Set(Q, H);
          }
        else
          {
            MH::RowType row;
            Q.InitializeRow(row);
            for (r=0; r<Q.Height(); ++r)
              {
                int c;
                for (c=mu; c<Q.Width(); ++c)
                  row(0,c)=Q(r,c);
                for (c=mu; c<Q.Width(); ++c)
                  {
                    MR::Number temp=0;
                    for (int i=mu; i<Q.Height(); ++i)
                      temp += H(i, c)*row(0,i);
                    Q(r,c)=temp;
                  }
              }
          }
        // Q should be orthogonal
        //  std::cout << "mu = " << mu << " Q =\n";
        //  Print(Q, std::cout);
      }
  }

// computes R = (1 - beta * u * transpose(u)) * R
// where u_0 .. u_(first-1) are presumed zero
template <class MR, class MU>
void HouseholderPreMult(MR & R, const MU & u, MU::Number beta, int first)
  {
    assert(IsSquare(R));
    assert(R.Height()==u.Height());
    assert(u.Width()==1);
    MU::TransposeType tur;
    u.InitializeTranspose(tur);
    // set tur to beta*transpose(u)*R
    int c, r;
    for (c=0; c<R.Width(); ++c)
      {
        MU::Number temp=0;
        for (r=first; r<u.Height(); ++r)
          temp += u(r, 0) * R(r, c);
        tur(0, c) = beta*temp;
      }
    // subtract u * tur from R
    for (r=first; r<R.Height(); ++r)
      for (c=0; c<R.Width(); ++c)
        R(r,c) -= u(r,0)*tur(0,c);
  }

// computes Q = Q * (1 - beta * u * transpose(u))
// where u_0 .. u_(first-1) are presumed zero
template <class MQ, class MU>
void HouseholderPostMult(MQ & Q, const MU & u, MU::Number beta, int first)
  {
    assert(IsSquare(Q));
    assert(Q.Height()==u.Height());
    assert(u.Width()==1);
    MU::ColumnType bqu;
    u.InitializeColumn(bqu);

    // set bqu to beta*Q*u
    int c, r;
    for (r=0; r<Q.Height(); ++r)
      {
        MU::Number temp=0;
        for (c=first; c<Q.Width(); ++c)
          temp += Q(r, c)*u(c, 0);
        bqu(r, 0) = beta*temp;
      }
    // subtract bqu*transpose(u) from Q
    for (r=0; r<Q.Height(); ++r)
      for (c=first; c<Q.Width(); ++c)
        Q(r,c) -= bqu(r,0)*u(c,0);
  }

// Decomposes R into QR with inverse(Q)=transpose(Q)
// and R upper triangular ** overwrites R **
// Modification suggested by John Strain
template <class MQ, class MR>
void QRDecomposition(MQ & Q, MR & R)
  { // All matrices should be distinct, square, and the same size
    assert(IsSquare(Q));
    assert(IsSquare(R));
    assert(Q.Width()==R.Width());
    assert(((void *)&Q) != ((void *)&R));
    SetToId(Q);
    for (int mu=0; mu<R.Width()-1; ++mu)
      {
        // a = bottom Height-mu rows of mu-th column

        // Compute norm a
        MR::Number norm_a=0;
        int r;
        for (r=mu; r<R.Height(); ++r)
          {
            MR::Number ar = R(r, mu);
            norm_a += ar*ar;
          }
        norm_a = sqrt(norm_a);

        // Compute beta, u
        MR::Number u0 = Abs(R(mu,mu))+norm_a;
        MR::Number beta = 1/(norm_a*u0);
        // u0 *= Sgn(R(mu,mu)); // replaced with:
        if (R(mu,mu)<0) u0 = -u0;
        MR::ColumnType u;
        R.InitializeColumn(u);
        u(mu,0)=u0;
        for (r=mu+1; r<R.Height(); ++r)
          u(r,0)=R(r,mu);


        // R = H * R
          {
            HouseholderPreMult(R, u, beta, mu);
          }
        // First mu+1 columns should be upper triangular
        //  std::cout << "mu = " << mu << " R =\n";
        //  Print(R, std::cout);

        // Q = Q * H
          {
            HouseholderPostMult(Q, u, beta, mu);
          }
        // Q should be orthogonal
        //  std::cout << "mu = " << mu << " Q =\n";
        //  Print(Q, std::cout);
      }
  }
}
#endif