#ifndef __ROOT_OF_SPLINE_1D_H
#define __ROOT_OF_SPLINE_1D_H

#include "template.h"
#include <list>
#include <vector>

namespace Root1D
  {
    typedef enum { Zero, Negative, Positive } SignResult;

    SignResult SignOfProduct(ScalarType a, ScalarType b);

    inline void AddResult(std::list<DomainType1d> &results, ScalarType x)
      {
        DomainType1d xc;
        xc(0,0)=x;
        results.push_back(xc);
      }

    class Hermite
      {
    public:
        ScalarType mX0;
        ScalarType mXd;
        ScalarType mY[2], mD[2];
        Hermite()
          { }
        Hermite(ScalarType x0, ScalarType xd, ScalarType y0, ScalarType y1, ScalarType d0, ScalarType d1);
        ScalarType BezierOrdinate(int i) const;
      };

    inline ScalarType Hermite::BezierOrdinate(int i) const
      {
        switch (i)
          {
            case 0:
              return mY[0];
            case 1:
              return mY[0] + mD[0]*mXd/3.;
            case 2:
              return mY[1] - mD[1]*mXd/3.;
            case 3:
              return mY[1];
          }
        assert(0);
        return 0;
      }


    class BezierUnit
      {
    public:
        ScalarType mY[4];
        void SetFromHermite(const Hermite &h);
         // t in 0..1
        void EvaluateAt(ScalarType t, ScalarType &y, ScalarType &d) const;
        ScalarType EvaluateAt(ScalarType t) const;
      };

    inline void BezierUnit::SetFromHermite(const Hermite &h)
      {
        for (int i=0; i<4; ++i)
          mY[i]=h.BezierOrdinate(i);
      }

    class BezierInterval
      {
    public:
        ScalarType mXd;
        ScalarType mX0;
        ScalarType mY[4];
        void SetFromHermite(const Hermite &h);

        // t in 0..1 corresponding to mX0..mX0+mXd
        void Subdivide(ScalarType t, BezierInterval &b);

        void EvaluateAt(ScalarType t, ScalarType &y, ScalarType &d) const;

        // 0 <= t0 <= t1 <= 1
        void Subsegment(ScalarType t0, ScalarType t1, Hermite &h) const;
      
      };

    inline void BezierInterval::SetFromHermite(const Hermite &h)
      {
        mXd = h.mXd;
        mX0 = h.mX0;

        for (int i=0; i<4; ++i)
          mY[i]=h.BezierOrdinate(i);
      }


    class BezierNthDegreeUnit
      {
        typedef BezierNthDegreeUnit self;

    protected:
        int mN;
        ScalarType *mY;

        void copy(const self & cpy);
        void destroy();

        // To evaluate (m j) (n k-j) / (m+n k) 
        //  = m! n! k! / (j! m-j! k-j! n-k+j! m+n!)
        static ScalarType BigChoose(int m, int n, int j, int k);

        static void Multiply(ScalarType *outY, int outN, 
                             ScalarType *lY,   int lN, 
                             ScalarType *rY,   int rN);

    public:
          // nth degree Bezier
        BezierNthDegreeUnit(int n)
              : mN(n), mY(new ScalarType[n+1])
          { assert(n>=0); }
          // constant Bezier of arbitrary degree
        BezierNthDegreeUnit(int n, ScalarType c)
              : mN(n), mY(new ScalarType[n+1])
          {
            assert(n>=0);
            for (int i=0; i<=n; ++i)
              mY[i]=c;
          }
          // linear Bezier, from (0,a) to (1,b)
        BezierNthDegreeUnit(int n, ScalarType a, ScalarType b) 
              : mN(n), mY(new ScalarType[n+1])
          {
            assert(n>=0);
            mY[0]=a;
            for (int i=1; i<=n; ++i)
               mY[i]=a + (b-a)*i/n;
          }
          // degree elevate cpy to nth degree Bezier
          // there is a formula, but it might just be
          // easier to multiply by the linear function 1 a few times
        //BezierNthDegreeUnit(int n, const self &cpy) 
        //      : mN(n), mY(new ScalarType[n])
        //  { 
        //    
        //  }
          // copy a Bezier
        BezierNthDegreeUnit(const self &cpy)
          { copy(cpy); }
        ~BezierNthDegreeUnit()
          { destroy(); }
        self & operator=(const self &cpy)
          {
            if (this != &cpy)
              {
                destroy();
                copy(cpy);
              }
            return *this;
          }

        void operator+=(const self & rhs);

        void operator-=(const self & rhs);

        void operator*=(ScalarType);

        void operator*=(const self & rhs);

        // performs a few extra copies...
        self operator*(const self & r);

        ScalarType *GetOrdinates()
          {
            return mY; 
          }
        void Rescale();
      };

    // Simplest bisection version of cubic root finding
    void Solve_Bisection(std::list<DomainType1d> &results, 
                        ScalarType x0, ScalarType y0, ScalarType d0,
                        ScalarType xd, ScalarType y1, ScalarType d1,
                        ScalarType epsilon);

    // Simplest bisection version of cubic root finding, Hermite version
    void Solve_Bisection(std::list<DomainType1d> &results, 
                         Hermite h, ScalarType epsilon);

    // Simplest bisection version of cubic root finding, Bezier version
    void Solve_Bisection(std::list<DomainType1d> &results, 
                         BezierInterval b,
                         ScalarType epsilon);

    inline bool LineSegmentIntersection(ScalarType y0, ScalarType y1, ScalarType xd, ScalarType &xresult)
      {
        if ((y0<0 && y1<0) || (y0>0 && y1>0))
          return false;
        if (y0==y1)
          xresult = 0;
        else
          xresult = xd * y0 / (y0-y1);
        return true;
      }

    // only used for slow Hull intersection code
    template<class ScalarType, class Vector>
    inline void FindMinMax(const Vector &v, ScalarType &mn, ScalarType &mx)
      {
        assert(!v.empty());
        mn=mx=v.front();
        Vector::const_iterator i=v.begin();
        assert(v.begin()!=v.end());
        for (++i; i!=v.end(); ++i)
          {
            if ((*i) > mx)
              mx = (*i);
            else if ((*i) < mn)
              mn = (*i);
          }
      }

    struct VerticalSegment
      {
        ScalarType mY[2]; // mY[0] <= mY[1]
      };

    void EnlargeExtents(std::vector<VerticalSegment> &hull, 
                        ScalarType epsilonR);

    // returns false if no intersection
    bool HullIntersection(const std::vector<VerticalSegment> &hull, 
                          ScalarType & minX, ScalarType & maxX);

    // returns false if no intersection
    // assumes hull has size at most 4
    bool HullIntersection_Fast(const std::vector<VerticalSegment>
                                &hull, ScalarType & minX, ScalarType & maxX);

    // Version of cubic root finding with quadratic convergence
    // Sherbrooke and Patrikalakis 1993 (CAGD 10 pp. 379-405)
    // Uses intersection of convex hull of control polygon with
    // x-axis to determine new interval
    // Simplest bisection version of cubic root finding, Bezier version
    void Solve_Quadratic(std::list<ScalarType> &results, 
                         BezierInterval b, 
                         ScalarType epsilon, ScalarType epsilonR);

    void Solve_Quadratic1D(std::list<ScalarType> &results, 
                           const BezierUnit &b, 
                           ScalarType epsilon, ScalarType epsilonR);

    void Solve_QuadraticBox(std::list<DomainType1d> &results, 
                           const BezierInterval &b,
                           ScalarType epsilon, ScalarType epsilonR);

    /* Only used by Solve_Quadratic_Safer 
    int CountIntersections(const std::vector<VerticalSegment> &hull);
    */
    // Safer version counts max number of root that can occur
    // in an interval and ensures that it won't be exceeded
    // (brute force, does not address cause; deflation method
    // below addresses the cause, but does not explicitly check
    // how many roots it generates)
    void Solve_Quadratic_Safer(std::list<ScalarType> &results, 
                         BezierInterval b, 
                         ScalarType epsilon, ScalarType epsilonR);


    // Like quadratic above, except works with arbitrary degree
    // and factors out any roots it finds
    void Solve_N(std::list<ScalarType> &results, int n, const ScalarType *y,
                         ScalarType x[2], 
                         ScalarType epsilon, ScalarType epsilonR);

    // Specializations of Solve_N for cubic Bezier splines
    inline void Solve_N_1D(std::list<ScalarType> &results, 
                           const BezierUnit &b, 
                           ScalarType epsilon, ScalarType epsilonR)
      {
        ScalarType x[2] = { 0, 1 };
        Solve_N(results, 3, b.mY, x, epsilon, epsilonR);
      }

    void Solve_N_P(std::list<ScalarType> &results, 
                           const BezierInterval &b, 
                           ScalarType epsilon, ScalarType epsilonR);

    void Solve_N_Box(std::list<DomainType1d> &results, 
                           const BezierInterval &b,
                           ScalarType epsilon, ScalarType epsilonR);
  }

#endif