#include "root2d.h"
#include "fixme.h"

namespace Root2D
  {
    // inTri should point to (n+1)*(n+2)/2 scalars
    // out should point to n*(n+1)/2 scalars
    
    void deCasteljauStep(const ScalarType *inTri, unsigned n, 
                         const ScalarType st[2], ScalarType *out)
      {
        assert(n>=1);
        for (int j=0; j<n; ++j)
          for (int i=0; i<n-j; ++i)
            {
              unsigned inoffset=((j*(3+(n<<1)-j))>>1)+i;
              unsigned inoffset2=(((j+1)*(2+(n<<1)-j))>>1)+i;
              unsigned outoffset=inoffset-j;
              out[outoffset] = (1-st[0]-st[1])*inTri[inoffset]+
                               st[0]*inTri[inoffset+1]+
                               st[1]*inTri[inoffset2];
            }

      }

    // inTri, out should each point to (n+1)*(n+2)/2 scalars
    
    void SubdivideBezierTriangle(const ScalarType *inTri, unsigned n, ScalarType st[3][2], ScalarType *out)
      {
        // the sum of the first n triangular numbers
        int buffsize=(n*(n+1)*(n+2))/6;
        boost::scoped_array<ScalarType>  buffers(new ScalarType[buffsize]);
        boost::scoped_array<ScalarType*> buffptr(new ScalarType *[n+1]);
        boost::scoped_array<int>         joffset(new int[n+1]);
        buffptr[0]=const_cast<ScalarType *>(inTri); // won't be written to
                                               // since loop reads from bufptr[i]
                                               // and writes to bufptr[i+1]
        // buffptr[1]=& buffers[0];
        // buffptr[2]=& buffers[n*(n+1)/2];
        // buffptr[3]=& buffers[n*(n+1)/2+(n-1)*n/2];
        // ...
        // buffptr[n]=& buffers[buffsize-1];
        int i, j, k;
        for (i=1; i<=n; ++i)
          {
            j=n-i+1;
            buffptr[i] = & buffers[ buffsize - (j*(j+1)*(j+2))/6 ];
          }

        for (j=0; j<=n; ++j)
          joffset[j] = (j*((n<<1)+3-j))>>1;

        i=0; 
        while (1)
          {
            j=0;
            while (1)
              {
                k=0;
                while (1)
                  {
                    deCasteljauStep(buffptr[i+j+k], n-i-j-k, st[0], buffptr[i+j+k+1]);
                    ++k;
                    if (i+j+k==n)
                      {
                        // buffptr[n] points to an array of size 1
                        out[joffset[j]+i] = * buffptr[n];
                        break;
                      }
                  }
                deCasteljauStep(buffptr[i+j], n-i-j, st[2], buffptr[i+j+1]);
                ++j;
                if (i+j==n)
                  {
                    // buffptr[n] points to an array of size 1
                    out[joffset[j]+i] = * buffptr[n];
                    break;
                  }
              }
            deCasteljauStep(buffptr[i], n-i, st[1], buffptr[i+1]);
            ++i;
            if (i==n)
              {
                // buffptr[n] points to an array of size 1
                out[i] = * buffptr[n];
                break;
              }
          }
      }

    //     2         5
    // a/b         c 3 4
    //     0   1     0 1 2

    
    void MultiplyBezierTriangle(
            const TriangularLinearUnit  & a, 
            const TriangularLinearUnit  & b, 
            TriangularQuadraticUnit  & c)
      {
        c.mY[0]=a.mY[0]*b.mY[0];
        c.mY[1]=(a.mY[0]*b.mY[1]+a.mY[1]*b.mY[0])/2;
        c.mY[2]=a.mY[1]*b.mY[1];

        c.mY[3]=(a.mY[0]*b.mY[2]+a.mY[2]*b.mY[0])/2;
        c.mY[4]=(a.mY[1]*b.mY[2]+a.mY[2]*b.mY[1])/2;

        c.mY[5]=a.mY[2]*b.mY[2];
      }

    //   2         5        9
    // a         b 3 4    c 7 8
    //   0   1     0 1 2    4 5 6
    //                      0 1 2 3

    
    void MultiplyBezierTriangle(
            const TriangularLinearUnit  & a, 
            const TriangularQuadraticUnit  & b, 
            TriangularCubicUnit  & c)
      {
        c.mY[0]=a.mY[0]*b.mY[0];
        c.mY[1]=(2*a.mY[0]*b.mY[1]+a.mY[1]*b.mY[0])/3;
        c.mY[2]=(2*a.mY[1]*b.mY[1]+a.mY[0]*b.mY[2])/3;
        c.mY[3]=a.mY[1]*b.mY[2];

        c.mY[4]=(2*a.mY[0]*b.mY[3]+a.mY[2]*b.mY[0])/3;
        c.mY[5]=(a.mY[0]*b.mY[4]+a.mY[1]*b.mY[3]+a.mY[2]*b.mY[1])/3;
        c.mY[6]=(2*a.mY[1]*b.mY[4]+a.mY[2]*b.mY[2])/3;

        c.mY[7]=(2*a.mY[2]*b.mY[3]+a.mY[0]*b.mY[5])/3;
        c.mY[8]=(2*a.mY[2]*b.mY[4]+a.mY[1]*b.mY[5])/3;

        c.mY[9]=a.mY[2]*b.mY[5];
        // check to see if I missed any terms:
        // a b  0 1 1 2 2 3 4 4 5 5 5 6 6 7 7 8 8 9 c
        // 0 0  x                                  
        // 0 1    x                                
        // 0 2          x                          
        // 0 3              x                      
        // 0 4                  x                  
        // 0 5                              x      
        // 1 0      x                              
        // 1 1        x                            
        // 1 2            x                        
        // 1 3                    x                
        // 1 4                        x            
        // 1 5                                  x  
        // 2 0                x                    
        // 2 1                      x              
        // 2 2                          x          
        // 2 3                            x        
        // 2 4                                x    
        // 2 5                                    x
      }


    
    void DegreeRaise(            
            const TriangularQuadraticUnit  & a, 
            TriangularCubicUnit  & b)
      {
        b.mY[0]=a.mY[0];
        b.mY[1]=(2*a.mY[1]+a.mY[0])/3;
        b.mY[2]=(2*a.mY[1]+a.mY[2])/3;
        b.mY[3]=a.mY[2];

        b.mY[4]=(2*a.mY[3]+a.mY[0])/3;
        b.mY[5]=(a.mY[1]+a.mY[3]+a.mY[4])/3;
        b.mY[6]=(2*a.mY[4]+a.mY[2])/3;

        b.mY[7]=(2*a.mY[3]+a.mY[5])/3;
        b.mY[8]=(2*a.mY[4]+a.mY[5])/3;

        b.mY[9]=a.mY[5];
      }

    
    void LinearApprox(
            const TriangularQuadraticUnit  & in,
            TriangularLinearUnit & out)
      { // could do better, but this is easy
        out.mY[0]=in.mY[0];
        out.mY[1]=in.mY[2];
        out.mY[2]=in.mY[5];
      }

    
    void LinearApprox(
            const TriangularCubicUnit  & in,
            TriangularLinearUnit & out)
      { // could do better, but this is easy
        out.mY[0]=in.mY[0];
        out.mY[1]=in.mY[3];
        out.mY[2]=in.mY[9];
      }

    
    int IntersectLinearFactors
            (const TriangularLinearUnit & l1,
             const TriangularLinearUnit & l2,
             ScalarType *intersection, ScalarType epsilonR)
      { // A X = B, ie.
        // [ ax1 ay1 ] [ x ] = [ b1 ]
        // [ ax2 ay2 ] [ y ] = [ b2 ]
        // where l1 = ax1 x + ay1 y - b1
        ScalarType ax1 = l1.mY[1] - l1.mY[0],
                   ay1 = l1.mY[2] - l1.mY[0],
                   b1  = -l1.mY[0];
        ScalarType ax2 = l2.mY[1] - l2.mY[0],
                   ay2 = l2.mY[2] - l2.mY[0],
                   b2  = -l2.mY[0];
        ScalarType det = ax1*ay2-ax2*ay1;
        if (Matrix::Abs(det)>epsilonR)
          {
            intersection[0] = (b1*ay2-b2*ay1)/det;
            intersection[1] = (ax1*b2-ax2*b1)/det;
            if (intersection[0]<0 || intersection[1]<0 ||
                intersection[0]+intersection[1]>1)
              return 0;
            return 1;
          }
        return 0;
      }

    // works in st space
    
    ScalarType TriangularQuadraticUnit ::EvaluateAt
          (const ScalarType st[2]) const
      {
        ScalarType temp1[3];
        ScalarType y;
        deCasteljauStep(mY, 2, st, temp1);
        deCasteljauStep(temp1, 1, st, &y);
        return y;
      }

    
    bool SameSign(const TriangularLinearUnit  & lin,
                         ScalarType epsilonR)
      {
        if (lin.mY[0]>epsilonR)
          return lin.mY[1]>epsilonR && lin.mY[2]>epsilonR;
        else if (lin.mY[0]<-epsilonR)
          return lin.mY[1]<-epsilonR && lin.mY[2]<-epsilonR;
        else
          return false;
      }

    
    bool SameSign(const TriangularQuadraticUnit  & q,
                         ScalarType epsilonR)
      {
        if (q.mY[0]>epsilonR)
          {
            for (int j=1; j<6; ++j)
              if (q.mY[j]<=epsilonR)
                return false;
            return true;
          }
        else if (q.mY[0]<-epsilonR)
          {
            for (int j=1; j<6; ++j)
              if (q.mY[j]>=-epsilonR)
                return false;
            return true;
          }
        else
          return false;
      }

    
    bool SameSign(const TriangularCubicUnit  & c,
                         ScalarType epsilonR)
      {
        if (c.mY[0]>epsilonR)
          {
            for (int j=1; j<10; ++j)
              if (c.mY[j]<=epsilonR)
                return false;
            return true;
          }
        else if (c.mY[0]<-epsilonR)
          {
            for (int j=1; j<10; ++j)
              if (c.mY[j]>=-epsilonR)
                return false;
            return true;
          }
        else
          return false;
      }

    // works in st space
    
    void TriangularCubicUnit ::EvaluateAt
          (const ScalarType st[2], ScalarType &y, ScalarType gradient[2]) const
      {
        ScalarType temp1[3];
        ScalarType temp2[6];
        deCasteljauStep(mY, 3, st, temp2);
        deCasteljauStep(temp2, 2, st, temp1);
        deCasteljauStep(temp1, 1, st, &y);
        // should be able to figure out gradient from temp1 and st
        gradient[0] = 3*(temp1[1]-temp1[0]);
        gradient[1] = 3*(temp1[2]-temp1[0]);
      }

    
    ScalarType TriangularCubicUnit ::EvaluateAt
                (const ScalarType st[2]) const
      {
        ScalarType temp1[3];
        ScalarType temp2[6];
        ScalarType y;
        deCasteljauStep(mY, 3, st, temp2);
        deCasteljauStep(temp2, 2, st, temp1);
        deCasteljauStep(temp1, 1, st, &y);
        return y;
      }

    // output two triangles corresponding to the axis-aligned rectangle of st
    
    void TriangularCubicUnit ::Subdivide(ScalarType st[2][2], thistype out[2]) const
      {
        ScalarType newst1[3][2] =
          { { st[0][0], st[0][1] },
            { st[1][0], st[0][1] },
            { st[0][0], st[1][1] } };

        SubdivideBezierTriangle(mY, 3, newst1, out[0].mY);
        // can probably be more efficient about generating ordinates
        // than:
        // ScalarType newst2[3][2] =
        //  { { st[1][0], st[1][1] },
        //    { st[0][0], st[1][1] },
        //    { st[1][0], st[0][1] } };
        // SubdivideBezierTriangle(mY, 3, newst2, out[1].mY);
        // for second triangle: for example, shares four points with first triangle
        out[1].mY[3] = out[0].mY[9];
        out[1].mY[6] = out[0].mY[8];
        out[1].mY[8] = out[0].mY[6];
        out[1].mY[9] = out[0].mY[3];

        // in fact, it is not that hard: just need to figure out basis vectors
        // of first in terms of basis vectors of second
        // i.e. (1-s-t)^3 = (1-x-y)^3 = (1-(1-s')-(1-t'))^3 = (-1+s'+t')^3 = -(1-s'-t')^3
        // (and in general: (1-s-t) -> -(1-s'-t')
        //                  s -> 1 - s' = (1-s'-t') + t'
        //                  t -> 1 - t' = (1-s'-t') + s' )
        // 3s(1-s-t)^2 = 3(1-s')(1-s'-t')^2 = 3(1-s'-t')^3 + 3t'(1-s'-t')^2
        out[1].mY[0] = -   out[0].mY[0] + 3*out[0].mY[1] - 3*out[0].mY[2]
                       +   out[0].mY[3] + 3*out[0].mY[4] - 6*out[0].mY[5]
                       + 3*out[0].mY[6] - 3*out[0].mY[7] + 3*out[0].mY[8]
                       +   out[0].mY[9];
        out[1].mY[1] = out[0].mY[4] - 2*out[0].mY[5] + out[0].mY[6] 
                   - 2*out[0].mY[7] + 2*out[0].mY[8] + out[0].mY[9];
        out[1].mY[2]= -out[0].mY[7] +   out[0].mY[8] + out[0].mY[9];
        out[1].mY[4] = out[0].mY[1] - 2*out[0].mY[2] + out[0].mY[3]
                   - 2*out[0].mY[5] + 2*out[0].mY[6] + out[0].mY[8];
        out[1].mY[5]= -out[0].mY[5] +   out[0].mY[6] + out[0].mY[8];
        out[1].mY[7]= -out[0].mY[2] +   out[0].mY[3] + out[0].mY[6];
      }

    // given line segment (pts[0][0], pts[0][1])-(pts[1][0], pts[1][1])
    // determines restriction of function to line segment in st space
    // returned as an Hermite spline 
    
    void TriangularCubicUnit ::CubicFromInterval(ScalarType pts[2][2], Root1D::Hermite  &h) const
      {
        ScalarType gradient[2][2];
        EvaluateAt(pts[0], h.mY[0], gradient[0]);
        EvaluateAt(pts[1], h.mY[1], gradient[1]);
        h.mX0=0;
        h.mXd=1;
        // figure out h.mD[*] from pts and gradient
        // v = pts[1]-pts[0]
        ScalarType v[2];
        v[0]=pts[1][0]-pts[0][0];
        v[1]=pts[1][1]-pts[0][1];
        // l = sqrt(Dot(v,v))
        ScalarType length=sqrt(v[0]*v[0]+v[1]*v[1]);
        if (length>0)
          // D[i] = Dot(gradient[i],v)/l;
          for (int i=0; i<2; ++i)
            h.mD[i] = (gradient[i][0]*v[0]+gradient[i][1]*v[1])/length;
        else
          h.mD[0] = h.mD[1] = 0;
      }

    
    void TriangularCubicUnit ::CubicFromSide(int side, Root1D::BezierUnit  &b) const
      {
        switch (side)
          {
            case 0:
              b.mY[0]=mY[0];
              b.mY[1]=mY[1];
              b.mY[2]=mY[2];
              b.mY[3]=mY[3];
              break;
            case 1:
              b.mY[0]=mY[3];
              b.mY[1]=mY[6];
              b.mY[2]=mY[8];
              b.mY[3]=mY[9];
              break;
            case 2:
              b.mY[0]=mY[9];
              b.mY[1]=mY[7];
              b.mY[2]=mY[4];
              b.mY[3]=mY[0];
              break;
            default:
              assert(0);
          }
      }

    
    inline void TriangularCubicUnit ::Cycle3(int a, int b, int c)
      {
        ScalarType temp = mY[a];
        mY[a]=mY[c];
        mY[c]=mY[b];
        mY[b]=temp;
      }

        //  ^ 9
        //  | 7 8
        //  t 4 5 6
        //    0 1 2 3
        //      s-->

    
    void TriangularCubicUnit ::RotateClockwise()
      {
        Cycle3(0, 9, 3);
        Cycle3(1, 7, 6);
        Cycle3(4, 8, 2);
      }

    
    void TriangularCubic ::RotateClockwise()
      {
        TriangularCubicUnit ::RotateClockwise();
        DomainType2d temp(mP[0]);
        mP[0]=mP[1];
        mP[1]=mP[2];
        mP[2]=temp;
      }

    
    void TriangularCubicUnit ::VerticalExtents(std::vector<Root1D::VerticalSegment  > &extents, STEnum st) const
      {
        Root1D::VerticalSegment  toadd;
        if (st==S)
          {
            InitExtents(toadd, 0);
            UpdateExtents(toadd, 4);
            UpdateExtents(toadd, 7);
            UpdateExtents(toadd, 9);
            extents.push_back(toadd);
            InitExtents(toadd, 1);
            UpdateExtents(toadd, 5);
            UpdateExtents(toadd, 8);
            extents.push_back(toadd);
            InitExtents(toadd, 2);
            UpdateExtents(toadd, 6);
            extents.push_back(toadd);
            InitExtents(toadd, 3);
            extents.push_back(toadd);
          }
        else
          { // st==T
            InitExtents(toadd, 0);
            UpdateExtents(toadd, 1);
            UpdateExtents(toadd, 2);
            UpdateExtents(toadd, 3);
            extents.push_back(toadd);
            InitExtents(toadd, 4);
            UpdateExtents(toadd, 5);
            UpdateExtents(toadd, 6);
            extents.push_back(toadd);
            InitExtents(toadd, 7);
            UpdateExtents(toadd, 8);
            extents.push_back(toadd);
            InitExtents(toadd, 9);
            extents.push_back(toadd);
          }
      }


    
    void TriangularCubicUnit ::
    PartialDerivativeS(TriangularQuadraticUnit  & out) const
      {
        out.mY[0] = 3*(mY[1]-mY[0]);
        out.mY[1] = 3*(mY[2]-mY[1]);
        out.mY[2] = 3*(mY[3]-mY[2]);

        out.mY[3] = 3*(mY[5]-mY[4]);
        out.mY[4] = 3*(mY[6]-mY[5]);

        out.mY[5] = 3*(mY[8]-mY[7]);
      }

    
    void TriangularCubicUnit ::
    PartialDerivativeT(TriangularQuadraticUnit  & out) const
      {
        out.mY[0] = 3*(mY[4]-mY[0]);
        out.mY[1] = 3*(mY[5]-mY[1]);
        out.mY[2] = 3*(mY[6]-mY[2]);

        out.mY[3] = 3*(mY[7]-mY[4]);
        out.mY[4] = 3*(mY[8]-mY[5]);

        out.mY[5] = 3*(mY[9]-mY[7]);
      }

    /*
        // given line segment (pts[0][0], pts[0][1])-(pts[1][0], pts[1][1])
        // find roots of this function restricted to that segment
        // returns false if function is identically 0 on the interval
    
    bool RootsInInterval(const TriangularCubicUnit  &patch, 
                        ScalarType pts[2][2], ScalarType epsilon, 
                        std::list<ScalarType> & roots)
      {
        Root1D::Hermite  h;
        patch.CubicFromInterval(pts, h);
        Root1D::BezierInterval  b;
        b.SetFromHermite(h);
        if (BezierIsZero(b, epsilon))
          return false;
        Root1D::Solve_Quadratic1D(roots, b, epsilon);
        return true;
      }
      */

        // output two triangles corresponding to the axis-aligned rectangle of st
    
    void TriangularQuadraticUnit ::Subdivide(ScalarType st[2][2], thistype out[2]) const
      {
        ScalarType newst1[3][2] =
          { { st[0][0], st[0][1] },
            { st[1][0], st[0][1] },
            { st[0][0], st[1][1] } };

        SubdivideBezierTriangle(mY, 2, newst1, out[0].mY);
        // see TriangularCubicUnit::Subdivide above
        out[1].mY[0] =  out[0].mY[0] - 2*out[0].mY[1] + out[0].mY[2]
                    - 2*out[0].mY[3] + 2*out[0].mY[4] + out[0].mY[5];
        out[1].mY[1] = -out[0].mY[3] +   out[0].mY[4] + out[0].mY[5];
        out[1].mY[2] =  out[0].mY[5];
        out[1].mY[3] = -out[0].mY[1] +   out[0].mY[2] + out[0].mY[4];
        out[1].mY[4] =  out[0].mY[4];
        out[1].mY[5] =  out[0].mY[2];
      }

    
    inline void TriangularQuadraticUnit ::InitExtents(Root1D::VerticalSegment  &segment, int index) const
      {
        segment.mY[0]=segment.mY[1]=mY[index];
      }

    
    inline void TriangularQuadraticUnit ::UpdateExtents(Root1D::VerticalSegment  &segment, int index) const
      {
        ScalarType y=mY[index];
        if (y<segment.mY[0])
          segment.mY[0] = y;
        else if (y>segment.mY[1])
          segment.mY[1] = y;
      }

    
    void TriangularQuadraticUnit ::VerticalExtents(std::vector<Root1D::VerticalSegment  > &extents, STEnum st) const
      {
        Root1D::VerticalSegment  toadd;
        if (st==S)
          {
            InitExtents(toadd, 0);
            UpdateExtents(toadd, 3);
            UpdateExtents(toadd, 5);
            extents.push_back(toadd);
            InitExtents(toadd, 1);
            UpdateExtents(toadd, 4);
            extents.push_back(toadd);
            InitExtents(toadd, 2);
            extents.push_back(toadd);
          }
        else
          { // st==T
            InitExtents(toadd, 0);
            UpdateExtents(toadd, 1);
            UpdateExtents(toadd, 2);
            extents.push_back(toadd);
            InitExtents(toadd, 3);
            UpdateExtents(toadd, 4);
            extents.push_back(toadd);
            InitExtents(toadd, 5);
            extents.push_back(toadd);
          }
      }

    // works in st space
    
    void TriangularQuadraticUnit ::EvaluateAt
          (const ScalarType st[2], ScalarType &y, ScalarType gradient[2]) const
      {
        ScalarType temp1[3];
        deCasteljauStep(mY, 2, st, temp1);
        deCasteljauStep(temp1, 1, st, &y);
        // should be able to figure out gradient from temp1 and st
        gradient[0] = 2*(temp1[1]-temp1[0]);
        gradient[1] = 2*(temp1[2]-temp1[0]);
      }

    // MISC HELPER FUNCTION for SimultaneousSolve(...) ---------------------

    
    struct STRectangle
      {
        ScalarType st[2][2];
      };

    //
    //bool CommonIntersection(const TriangularCubicUnit  &cubic,
    //                        const TriangularQuadraticUnit  &quadratic,
    //                        STRectangle  &output);


    template<class ScalarType, class Patch1Type, class Patch2Type>
    bool CommonIntersection(const Patch1Type &patch1,
                            const Patch2Type &patch2,
                            ScalarType epsilonR,
                            STRectangle  &output)
      {
        std::vector<Root1D::VerticalSegment  > extents;
        for (int st=0; st<2; ++st)
          {
            patch1.VerticalExtents(extents, static_cast<STEnum>(st));
            Root1D::EnlargeExtents(extents, epsilonR);
            if (! Root1D::HullIntersection_Fast(extents, output.st[0][st], 
                                                output.st[1][st]))
              return false;
            extents.clear();

            ScalarType mn, mx;
            patch2.VerticalExtents(extents, static_cast<STEnum>(st));
            Root1D::EnlargeExtents(extents, epsilonR);
            if (! Root1D::HullIntersection_Fast(extents, mn, mx))
              return false;
            if (mn>output.st[0][st])
              output.st[0][st]=mn;
            if (mx<output.st[1][st])
              output.st[1][st]=mx;
            extents.clear();

            if (output.st[0][st] > output.st[1][st])
              return false;
          }
        return true;
      }

    // SimultaneousSolve(...) ----------------------------------------------

    // goal is to find points where Cubic and Quadratic are both 0
    // Implements Sherbrooke1993 modified for triangular cubic patches
    template<class ScalarType, class Patch1Type, class Patch2Type>
    void SimultaneousSolveTemplate(const Patch1Type &patch1, 
                           const Patch2Type &patch2, 
                           ScalarType epsilon, ScalarType epsilonR,
                           std::list<STPoint  > &ret)
      { // issue: how do we avoid duplicate roots?
        STRectangle  rect;
        if (! CommonIntersection(patch1, patch2, epsilonR, rect))
          return;
        //bool filterresults=(rect.st[1][0]+rect.st[1][1] > 1);
        //struct ResultType
        //  {
        //    ScalarType st[2];
        //    DomainType2d domain;
        //  };
        //std::list<ResultType> results;

        std::list<STRectangle  > worklist;
        worklist.push_back(rect);
        while (! worklist.empty())
          {
            rect=worklist.back();
            worklist.pop_back();
            if (rect.st[0][0]+rect.st[0][1] > 1)
              continue;
            Patch1Type::UnitType sub1[2];
            Patch2Type           sub2[2];
            patch1.Subdivide(rect.st, sub1);
            patch2.Subdivide(rect.st, sub2);
            STRectangle  rects[2];
            bool nonempty[2];
            int i,j;
            for (i=0; i<2; ++i)
              nonempty[i]=CommonIntersection(sub1[i], sub2[i], epsilonR, rects[i]);
            if (nonempty[1])
              {
                // convert from inverted coordinate system
                for (j=0; j<2; ++j)
                  {
                    for (i=0; i<2; ++i)
                      rects[1].st[i][j] = 1-rects[1].st[i][j];
                    std::swap(rects[1].st[0][j], rects[1].st[1][j]);
                  }
                if (nonempty[0])
                  { // check for overlap
                    if (rects[0].st[0][0] <= rects[1].st[1][0] &&
                        rects[0].st[0][1] <= rects[1].st[1][1] &&
                        rects[0].st[1][0] >= rects[1].st[0][0] &&
                        rects[0].st[1][1] >= rects[1].st[0][1])
                      { // they overlap, so simplify to their union
                        for (int st=0; st<2; ++st)
                          {
                            if (rects[0].st[0][st]>rects[1].st[0][st])
                              rects[0].st[0][st]=rects[1].st[0][st];
                            if (rects[0].st[1][st]<rects[1].st[1][st])
                              rects[0].st[1][st]=rects[1].st[1][st];
                          }
                        nonempty[1]=false;
                      }
                  }
              }
            for (i=0; i<2; ++i)
              if (nonempty[i])
                { 
                  int subdivisions[2];
                  ScalarType stlens[2]= { rects[i].st[1][0]-rects[i].st[0][0], 
                                          rects[i].st[1][1]-rects[i].st[0][1] };
                  bool smallenough=true;
                  for (j=0; j<2; ++j)
                    { // determine if we are going to need to subdivide
                      if (stlens[j] > .8)
                        subdivisions[j]=2;
                      else
                        subdivisions[j]=1;
                      // convert from local to global st coordinates
                      ScalarType scalefact = rect.st[1][j]-rect.st[0][j];
                      stlens[j]*=scalefact;
                      rects[i].st[0][j]*=scalefact;
                      rects[i].st[0][j]+=rect.st[0][j];
                      // can have a box sqrt(2)*epsilon on a side
                      // and still be within epsilon of the root
                      // (since we can choose the center of the box)
                      if (stlens[j]<1.4*epsilon)
                        { // don't shrink to smaller than epsilon
                          // otherwise roundoff might cause us to lose
                          // the root
                          subdivisions[j]=1;
                          ScalarType newstlen = epsilon;
                          ScalarType enlargeby = (newstlen-stlens[j])*.5;
                          if (enlargeby>0)
                            {
                              stlens[j]=newstlen;
                              rects[i].st[0][j]-=enlargeby;
                            }
                        }
                      else
                        {
                          smallenough=false;
                        }
                    }

                  // if small enough add to ret (but only if s+t<=1)
                  if (smallenough)
                    {
                      // could assume p are well approximated
                      // by planes and find their common intersection with 0
                      // (this would noticibly improve the accuracy)
                      // but then would have to deal with many special cases
                      // (like parallel intersection lines, intersection
                      // outside rects[i], collinear intersection lines, etc.)
#if 0
                      // original code
                      STPoint  result;
                      result.st[0]=rects[i].st[0][0]+stlens[0]/2;
                      result.st[1]=rects[i].st[0][1]+stlens[1]/2;
                      if (result.st[0]+result.st[1] <= 1)
                        ret.push_back(result);
#else
                      STRectangle  rect2;
                      rect2.st[0][0]=rects[i].st[0][0];
                      rect2.st[0][1]=rects[i].st[0][1];
                      rect2.st[1][0]=rects[i].st[0][0]+stlens[0];
                      rect2.st[1][1]=rects[i].st[0][1]+stlens[1];
                      patch1.Subdivide(rect2.st, sub1);
                      patch2.Subdivide(rect2.st, sub2);
                      TriangularLinearUnit  lin[4];
                      ScalarType answer[4];
                      int num=0;
                      for (int which=0; which<2; ++which)
                        {
                          LinearApprox(sub1[which], lin[2*which]);
                          LinearApprox(sub2[which], lin[2*which+1]);
                          int hasintersect=IntersectLinearFactors
                                            (lin[2*which], lin[2*which+1], 
                                             & answer[num*2],
                                             epsilonR);
                          if (hasintersect)
                            {
                              if (which)
                                {
                                  answer[num*2]=1-answer[num*2];
                                  answer[num*2+1]=1-answer[num*2+1];
                                }
                              ++num;
                            }
                        }
                      if (num)
                        {
                          if (num>1)
                            {
                              answer[0]+=answer[2];
                              answer[1]+=answer[3];
                              answer[0]*=0.5;
                              answer[1]*=0.5;
                            }
                          STPoint  result;
                          result.st[0]=rects[i].st[0][0]+answer[0]*stlens[0];
                          result.st[1]=rects[i].st[0][1]+answer[1]*stlens[1];
                          if (result.st[0]+result.st[1] <= 1)
                            ret.push_back(result);
                        }
                      //else if ((! SameSign(sub1[0]) && ! SameSign(sub2[0])) ||
                      //         (! SameSign(sub1[1]) && ! SameSign(sub2[1])))
                      else if ((! SameSign(lin[0], epsilonR) && 
                                ! SameSign(lin[1], epsilonR)) ||
                               (! SameSign(lin[2], epsilonR) && 
                                ! SameSign(lin[3], epsilonR)))
                        {
                          STPoint  result;
                          result.st[0]=rects[i].st[0][0]+stlens[0]/2;
                          result.st[1]=rects[i].st[0][1]+stlens[1]/2;
                          if (result.st[0]+result.st[1] <= 1)
                            ret.push_back(result);
                        }
#endif
                    }
                  else
                    { // otherwise add rects[i] to the worklist,
                      // subdividing if too big
                      for (j=0; j<2; ++j)
                        stlens[j]/=subdivisions[j];
                      for (j=0; j<subdivisions[0]; ++j)
                        for (int k=0; k<subdivisions[1]; ++k)
                          {
                            STRectangle  toadd;
                            toadd.st[0][0]=rects[i].st[0][0]+stlens[0]*j;
                            toadd.st[0][1]=rects[i].st[0][1]+stlens[1]*k;
                            toadd.st[1][0]=rects[i].st[0][0]+stlens[0]*(j+1);
                            toadd.st[1][1]=rects[i].st[0][1]+stlens[1]*(k+1);
                            worklist.push_back(toadd);
                          }
                    }
                }
          }
      }

    
    void SimultaneousSolve(const TriangularCubic 
                                   &cubic,
                           const TriangularQuadraticUnit 
                                   &quadratic, 
                           ScalarType epsilon, ScalarType epsilonR,
                           std::list<STPoint  > &ret)
      {
        SimultaneousSolveTemplate(cubic, quadratic, epsilon, epsilonR, ret);
      }
    // alternate overload
    
    void SimultaneousSolve(const TriangularQuadratic 
                                   &quadratic1,
                           const TriangularQuadraticUnit
                                   &quadratic2,
                           ScalarType epsilon, ScalarType epsilonR,
                           std::list<STPoint  > &ret)
      {
        SimultaneousSolveTemplate(quadratic1, quadratic2, epsilon, epsilonR, ret);
      }
  }

