#ifndef __ROOT_OF_SPLINE_2D_H
#define __ROOT_OF_SPLINE_2D_H

#include "template.h"
#include <list>
#include <vector>
#include "matrix.h"
#include "root1d.h"
#include <boost/scoped_array.hpp>

namespace Root2D
  {

    // in should point to (n+1)*(n+2)/2 scalars
    // out should point to n*(n+1)/2 scalars
    void deCasteljauStep(const ScalarType *inTri, unsigned n, 
                          const ScalarType st[2], ScalarType *out);

    // in, out should each point to (n+1)*(n+2)/2 scalars
    void SubdivideBezierTriangle(const ScalarType *inTri, unsigned n, 
                                ScalarType st[3][2], ScalarType *out);

    typedef enum { S=0, T=1 } STEnum;

    class TriangularQuadraticUnit
      {
    protected:
        typedef TriangularQuadraticUnit thistype;
    public:
        typedef thistype UnitType;
        //  ^ 5
        //  | 3 4
        //  t 0 1 2
        //     s-->
        ScalarType mY[6];
        // output two triangles corresponding to the axis-aligned rectangle 
        // of st
        void Subdivide(ScalarType st[2][2], thistype out[2]) const;
    protected:
        inline void InitExtents(Root1D::VerticalSegment  
                                &segment, int index) const;
        inline void UpdateExtents(Root1D::VerticalSegment  
                                &segment, int index) const;
    public:
        void VerticalExtents(std::vector<Root1D::VerticalSegment> &extents, STEnum st) const;
        void EvaluateAt(const ScalarType st[2], ScalarType &y, 
                            ScalarType gradient[2]) const;
        ScalarType EvaluateAt(const ScalarType st[2]) const;

        void SetToConstant(ScalarType constant)
          { for (int i=0; i<6; ++i)
              mY[i] = constant; }

        void AddS(ScalarType s)
          { // s is zero on 0, 3, 5
            mY[1] += s/2;
            mY[4] += s/2;
            mY[2] += s;
          }

        void AddT(ScalarType t)
          { // t is zero on 0, 1, 2
            mY[3] += t/2;
            mY[4] += t/2;
            mY[5] += t;
          }

        void AddS2(ScalarType s2)
          { // s*s is zero on 0, 1, 3, 4, 5
            mY[2] += s2;
          }

        void AddT2(ScalarType t2)
          { // t*t is zero on 0, 1, 2, 3, 4
            mY[5] += t2;
          }
      };

    
    class TriangularQuadratic 
        : public TriangularQuadraticUnit 
      {
    public:
        typedef TriangularQuadraticUnit  UnitType;
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> STLocation;

        DomainType2d mP[3];
        DomainType2d STToDomain(const STLocation &in)
          {
            DomainType2d ret(mP[0]);
            Matrix::AddScaledDiff(ret, in(0,0), mP[1], mP[0]);
            Matrix::AddScaledDiff(ret, in(0,1), mP[2], mP[0]);
            return ret;
          }
      };


    
    class TriangularCubicUnit
      {
    protected:
        typedef TriangularCubicUnit  thistype;

    public:
        typedef thistype UnitType;
        //  ^ 9
        //  | 7 8
        //  t 4 5 6
        //    0 1 2 3
        //      s-->
        ScalarType mY[10];

        // works in st space
        void EvaluateAt(const ScalarType st[2], ScalarType &y, ScalarType gradient[2]) const;
        ScalarType EvaluateAt(const ScalarType st[2]) const;

        // output two triangles corresponding to the axis-aligned rectangle of st
        void Subdivide(ScalarType st[2][2], thistype out[2]) const;

        // given line segment (pts[0][0], pts[0][1])-(pts[1][0], pts[1][1])
        // determines restriction of function to line segment in st space
        // returned as an Hermite spline 
          // NOT USED YET
        void CubicFromInterval(ScalarType pts[2][2], Root1D::Hermite  &h) const;
        // cubic for a side of the triangle returned as a Bezier spline
        void CubicFromSide(int side, Root1D::BezierUnit  &b) const;

    protected:
        inline void Cycle3(int a, int b, int c);
    public:
        void RotateClockwise();

    protected:
        inline void InitExtents(Root1D::VerticalSegment  &segment, int index) const;
        inline void UpdateExtents(Root1D::VerticalSegment  &segment, int index) const;
    public:
        void VerticalExtents(std::vector<Root1D::VerticalSegment  > &extents, STEnum st) const;

        void PartialDerivativeS(TriangularQuadraticUnit  & 
                                    out) const;
        void PartialDerivativeT(TriangularQuadraticUnit  & 
                                    out) const;
      };

    
    inline void TriangularCubicUnit ::InitExtents(Root1D::VerticalSegment  &segment, int index) const
      {
        segment.mY[0]=segment.mY[1]=mY[index];
      }

    
    inline void TriangularCubicUnit ::UpdateExtents(Root1D::VerticalSegment  &segment, int index) const
      {
        ScalarType y=mY[index];
        if (y<segment.mY[0])
          segment.mY[0] = y;
        else if (y>segment.mY[1])
          segment.mY[1] = y;
      }

    
    class TriangularLinearUnit
      {
    protected:
        typedef TriangularLinearUnit  thistype;
    public:
        //  ^ 2
        //  |  
        //  t 0  1 
        //    s-->
        ScalarType mY[3];
      };

    
    int IntersectLinearFactors
            (const TriangularLinearUnit & l1,
             const TriangularLinearUnit & l2,
             ScalarType *intersection, ScalarType epsilonR);

    // c = a * b
    
    void MultiplyBezierTriangle(
            const TriangularLinearUnit  & a, 
            const TriangularLinearUnit  & b, 
            TriangularQuadraticUnit  & c);
    
    void MultiplyBezierTriangle(
            const TriangularLinearUnit  & a, 
            const TriangularQuadraticUnit  & b, 
            TriangularCubicUnit  & c);

    // out=in
    
    void DegreeRaise(            
            const TriangularQuadraticUnit  & in, 
            TriangularCubicUnit  & out);

    // out approx= in
    
    void LinearApprox(
            const TriangularQuadraticUnit  & in,
            TriangularLinearUnit & out);

    // out approx= in
    
    void LinearApprox(
            const TriangularCubicUnit  & in,
            TriangularLinearUnit & out);

    // -------- Not implemented yet --------
    // in1 should point to (n1+1)*(n1+2)/2 scalars
    // in2 should point to (n2+1)*(n2+2)/2 scalars
    // out should point to (n1+n2+1)*(n1+n2+2)/2 scalars
    // out = in1 * in2
    /*
      ( Sum_{i,j,k >= 0, i+j+k=na}    a(i,j,k) C(na, i, j, k) s^i t^j r^k ) *
      ( Sum_{i,j,k >= 0, i+j+k=nb}    b(i,j,k) C(nb, i, j, k) s^i t^j r^k ) =
      ( Sum_{i,j,k >= 0, i+j+k=na+nb} c(i,j,k) C(na+nb, i, j, k) s^i t^j r^k )
      Where c(i,j,k) =
      ( Sum_{ia,ja,ka >= 0, ia+ja+ka=na, ia<=i, ja<=j, ka<=k}  
          a(ia, ja, ka) C(na, ia, ja, ka) *
          b(i-ia, j-ja, k-ka) C(nb, i-ia, j-ja, k-ka) ) / C(na+nb, i, j, k)

    */
    //
    //void MultiplyBezierTriangle(const ScalarType *in1, unsigned n1, 
    //                            const ScalarType *in2, unsigned n2,
    //                            ScalarType *out);



        // BezierType==Root1D::BezierInterval 
    // or BezierType==Root1D::BezierUnit 
    template<class BezierType, class ScalarType>
    inline bool BezierIsZero(const BezierType &b, ScalarType epsilonR)
      {
        int i;
        for (i=0; i<4; ++i)
          if (Matrix::Abs(b.mY[i])>epsilonR)
            return false;
        return true;
      }

        // given line segment (pts[0][0], pts[0][1])-(pts[1][0], pts[1][1])
        // find roots of this function restricted to that segment
        // returns false if function is identically 0 on the interval
    /*      // NOT USED YET
    
    bool RootsInInterval(const TriangularCubicUnit  &patch, 
                        ScalarType pts[2][2], ScalarType epsilon, 
                        std::list<ScalarType> & roots);
    */

    
    class TriangularCubic : public TriangularCubicUnit 
      {
    public:
        typedef TriangularCubicUnit  UnitType;
        typedef Matrix::SmallMatrix<ScalarType, 1, 2> STLocation;

        DomainType2d mP[3];
        void RotateClockwise();
        DomainType2d STToDomain(const STLocation &in)
          {
            DomainType2d ret(mP[0]);
            Matrix::AddScaledDiff(ret, in(0,0), mP[1], mP[0]);
            Matrix::AddScaledDiff(ret, in(0,1), mP[2], mP[0]);
            return ret;
          }
      };

    
    struct STPoint
      {
        ScalarType st[2];
      };


    
    bool SameSign(const TriangularLinearUnit  & lin,
                         ScalarType epsilonR);

    
    bool SameSign(const TriangularQuadraticUnit  & q,
                         ScalarType epsilonR);

    
    bool SameSign(const TriangularCubicUnit  & c,
                         ScalarType epsilonR);

    // true if patch is <= epsilonR within epsilion of st
    template<class ScalarType, class PatchType>
    bool NearAZero(const PatchType & patch, const ScalarType st[2],
                    ScalarType epsilon, ScalarType epsilonR)
      {
        ScalarType rect[2][2];
        rect[0][0]=st[0]-epsilon;
        rect[0][1]=st[1]-epsilon;
        rect[1][0]=st[0]+epsilon;
        rect[1][1]=st[1]+epsilon;
        PatchType::UnitType sub[2];
        patch.Subdivide(rect, sub);
        int num=0;
        for (int which=0; which<2; ++which)
          {
            if (! SameSign(sub[which], epsilonR))
              return true;
          }
        return false;
      }



    // goal is to find points where Cubic and Quadratic are both 0
    // Implements Sherbrooke1993 modified for triangular cubic patches
    
    void SimultaneousSolve(const TriangularCubic 
                                   &cubic,
                           const TriangularQuadraticUnit 
                                   &quadratic, 
                           ScalarType epsilon, ScalarType epsilonR,
                           std::list<STPoint  > &ret);
    // alternate overload
    
    void SimultaneousSolve(const TriangularQuadratic 
                                   &quadratic1,
                           const TriangularQuadraticUnit
                                   &quadratic2,
                           ScalarType epsilon, ScalarType epsilonR,
                           std::list<STPoint  > &ret);

    
    class Contour
      {
    public:
        std::vector<DomainType2d> mControl;
        int mEndingSides[2];
        // -1 == closed loop
        // 0-2 == side
        // 3-5 == corner
      };


// To write:
// filter for simplifying a contour

    // Implements Grandine1997 modified for triangular cubic patches
    
    void FindContour(const TriangularCubic 
                     &cubic, ScalarType epsilon, ScalarType epsilonR,
                     std::list<Contour> & results);
  }

#endif